# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations
from collections import OrderedDict
import functools
import inspect
import os
import sys
from types import ModuleType
from typing import Callable

from contrast.utils.decorators import fail_quietly
from contrast.utils.libraries import get_module_distribution_metadata
from contrast.utils.stdlib_modules import is_stdlib_module
from contrast_vendor import wrapt

from contrast.agent.policy import patch_manager
from contrast.utils.object_utils import get_name

from contrast_vendor.wrapt import function_wrapper
from contrast_vendor import structlog as logging

from contrast.utils.string_utils import ensure_string
from contrast_vendor.wrapt import importer
from contrast_vendor.wrapt.importer import register_post_import_hook
import contextlib

logger = logging.getLogger("contrast")


def add_watermark(func):
    """
    Adds a "secret" attribute to patched function for debugging purposes.

    Do not rely on the existence of this attribute in agent source code.
    """
    with contextlib.suppress(Exception):
        func.__contrast__ = True
    return func


def wrap_and_watermark(orig_func, wrapper):
    # NOTE: adding a watermark here doesn't make a ton of sense anymore since
    # 1. The function wrapper always has a __wrapped__ attribute, which is a good watermark itself
    # 2. We can't apply the watermark to the wrapped function because in that
    # case it actually gets applied to the underlying object (which we don't want).
    return function_wrapper(add_watermark(wrapper))(orig_func)


def pack_self(instance: object | None, args: tuple) -> tuple:
    """Combines the instance and args into a single tuple. If instance is None, returns args."""
    return args if instance is None else (instance,) + args


def get_arg(args, kwargs, idx, kw=None, default=None):
    if kw and kw in kwargs:
        return kwargs[kw]

    if len(args) <= idx:
        return default

    return args[idx]


def set_arg(
    value: object, args, kwargs, idx, kw=None
) -> tuple[tuple, dict[str, object]]:
    """
    Set the value in the args or kwargs at the given index or keyword.

    If a value is not present in args or kwargs, it will be added as a kwarg.
    """
    if (kw and kw in kwargs) or idx >= len(args):
        kwargs[kw] = value
    else:
        mut_args = list(args)
        mut_args[idx] = value
        args = tuple(mut_args)
    return args, kwargs


def build_and_apply_patch(
    owner,
    attr_name,
    patch_builder,
    builder_args=None,
    owner_name=None,
    force=False,
):
    """
    Builds new patch using given builder and applies it to specified patch location

    :param owner: Module or class where patch will be applied
    :param loc_name: Fully specified name of module or class where patch will apply
    :param attr_name: Name of the method that is being patched/replaced
    :param patch_builder: Callback function used to build new patch
    :param builder_args: A tuple of positional args to be passed to `patch_builder`

    The `patch_builder` function must take at least two arguments:
        1. A pointer to the original function
        2. The patch policy associated with this patch (may be `None`)
    The `patch_builder` may accept additional positional arguments that are passed to
    this function as a tuple via `builder_args`.

    The `patch_builder` function must return a function that matches the argument
    signature of the original function. The returned function must call the original
    function and return the result.

    Not all patches will have policy. Some patch locations are used solely to apply
    proxies or do library analysis, and so no policy exists for those locations.
    Callers can indicate this by passing "" or None for `loc_name`, in which case no
    patch policy will be retrieved.
    """
    original_func = getattr(owner, attr_name)

    # In most cases, we don't want to patch a function that has already been patched.
    # Sometimes we do, though, so we have a `force` flag to allow it.
    # An examples of when we might want to force a patch is applying automatic
    # CommonMiddlewarePatch with route coverage patches.
    if patch_manager.is_patched(original_func) and not force:
        return

    from contrast.agent.policy import registry

    loc_name = owner_name if owner_name is not None else get_name(owner)

    patch_policy = (
        registry.get_policy_by_name(f"{loc_name}.{attr_name}") if loc_name else None
    )

    patch = patch_builder(original_func, patch_policy, *(builder_args or ()))

    patch_manager.patch(owner, attr_name, patch)

    func = add_watermark(patch_manager.as_func(getattr(owner, attr_name)))
    if hasattr(func, "__name__") and not isinstance(func, wrapt.FunctionWrapper):
        func.__name__ = ensure_string(attr_name)


THIRD_PARTY_SUPPORTED_VERSIONS = {
    "aiohttp": ((3, 7), (3, 10)),
    "aiohttp_session": ((2, 0), (2, 12)),
    "beaker": ((1, 0), (1, 13)),
    "bottle": ((0, 13), (0, 13)),
    "bottle_session": ((1, 0), (1, 0)),
    "cgi": ((2, 6), (2, 6)),
    "Crypto": ((3,), (3,)),  # pycryptodome
    "Cryptodome": ((3,), (3,)),  # pycryptodome
    "django": (
        (2, 2),
        (5, 2),
    ),  # official Django minimum support is 3.2, but the official DRF minimum version uses Django 2.2
    "enumfields": ((2, 0), (2, 1)),  # django-enumfields
    "falcon": ((3, 0), (4, 0)),
    "falcon_multipart": ((0, 1), (0, 2)),
    "fastapi": ((0, 71), (0, 121)),
    "flask": ((1, 1), (3, 1)),
    "genshi": ((0, 7), (0, 7)),
    "graphene": ((3, 4), (3, 4)),
    "httpx": ((0,), (0,)),
    "jinja2": ((2, 10), (3, 1)),  # min version required by our min version of flask
    "loguru": ((0, 7), (0, 7)),
    "lxml": ((4, 1), (5, 4)),
    "markupsafe": ((1, 0), (2, 1)),
    "mod_wsgi": ((4, 1), (5, 0)),
    "mysql": ((8, 0), (9, 4)),  # mysql-connector-python
    "openai": ((0, 27), (0, 28)),
    "pymysql": ((1, 0), (1, 1)),
    "psycopg2": ((2, 0), (2, 9)),
    "pymongo": ((4, 0), (4, 8)),
    "pyramid": ((1, 10), (2, 0)),
    "_pytest": (
        (0,),
        (100,),
    ),  # this is only used for our own unit tests, so we can make it broad.
    "quart": ((0, 15), (0, 20)),
    "requests": ((2, 4), (2, 32)),
    "rest_framework": ((3, 12), (3, 16)),  # drf
    "simplejson": ((3, 17), (3, 20)),
    "sqlalchemy": ((1,), (2,)),
    "starlette": (
        (0, 17),
        (0, 49),
    ),  # fastapi==0.71.0 requires starlette==0.17.1, fastapi==0.120.1 requires starlette<0.50.0
    "urllib3": ((1, 25), (2, 5)),
    "webob": ((1, 8), (1, 9)),
    "werkzeug": ((1, 0), (3, 1)),  # flask==1.1.* resolved to Werkzeug==1.0.1
    "yaml": ((5, 1), (6, 0)),  # PyYAML
}

ModulePatcher = Callable[[ModuleType], None]


def register_module_patcher(patcher: ModulePatcher, module_name: str):
    """
    Register a patcher that will be called with the module object when the named module is imported.

    If the named module has already been imported, the patcher is called immediately.
    """
    is_contrast_module = module_name.startswith(("contrast.", "contrast_vendor."))
    if (
        not is_stdlib_module(module_name)
        and not is_contrast_module
        and not is_versioned_patch(patcher)
    ):
        top_level_module = module_name.split(".")[0]
        if constraint := THIRD_PARTY_SUPPORTED_VERSIONS.get(top_level_module):
            patcher = versioned_patch(*constraint)(patcher)
        else:
            raise ValueError(
                f"Patch for non-stdlib module {module_name} must have a version_constraint"
            )

    register_post_import_hook(
        patcher,
        module_name,
    )


def unregister_module_patcher(module_name: str):
    """
    Unregister a patcher that was registered with `register_module_patcher`.
    """
    with importer._post_import_hooks_lock:
        importer._post_import_hooks.pop(module_name, None)


def is_versioned_patch(patch: object) -> bool:
    return hasattr(patch, "__version_constraint__")


class UnsupportedVersion(Exception):
    def __init__(self, module: ModuleType, version: str) -> None:
        package_or_name = getattr(module, "__package__", "") or module.__name__
        super().__init__(
            f"{package_or_name}=={version} is outside of Contrast's supported range."
        )


def _raise_if_testing(module: ModuleType, version: str) -> None:
    if os.environ.get("CONTRAST_TESTING"):
        raise UnsupportedVersion(module, version)


def versioned_patch(
    min: tuple | None = None, max: tuple | None = None
) -> Callable[[ModulePatcher], ModulePatcher]:
    """
    Decorator to restrict patch application to a specific version range.
    """
    if min is None and max is None:
        raise ValueError("Version range must be specified")

    if min and max and len(min) != len(max):
        raise ValueError("Version ranges must be the same length")

    def _versioned_patch_builder(patcher: ModulePatcher):
        @fail_quietly("Failed to apply versioned patch")
        def _versioned_patch(module: ModuleType):
            if dist_meta := get_module_distribution_metadata(module):
                version = dist_meta["Version"]
                version_info = tuple(int(v) for v in version.split(".")[:3])
                if min and version_info[: len(min)] < min:
                    # version is too low
                    _raise_if_testing(module, version)
                    return
                if max and version_info[: len(max)] > max:
                    # version is too high
                    _raise_if_testing(module, version)
                    return
                # version is just right
                patcher(module)
            else:
                raise ValueError(f"Failed to get distribution metadata for {module}")

        _versioned_patch.__version_constraint__ = (min, max)
        return _versioned_patch

    return _versioned_patch_builder


def get_loaded_modules() -> OrderedDict[str, ModuleType]:
    """
    Retrieves, filters, and sorts all loaded modules. Sorting keeps patching in a
    deterministic order.

    NOTE: This method gets called multiple times during the course of agent
    initialization. Ideally it would be called only once for PERF optimization,
    but because sys.modules is global to all threads, we can't guarantee its contents
    will be the same and that a race condition won't happen which would add modules
    across different threads.
    """
    filtered = OrderedDict()
    filtered.update(
        {
            name: module
            for name, module in sorted(sys.modules.items())
            if inspect.ismodule(module)
        }
    )

    return filtered


def is_patchable(obj):
    if inspect.ismodule(obj):
        return False
    if inspect.isclass(obj):
        return True
    if type(obj) is functools.partial:
        return True

    # cython methods look like unpatchable builtins, but they can be patched normally
    # an example of this is lxml.etree.fromstring
    # for additional info, see https://groups.google.com/forum/#!topic/cython-users/v5dXFOu-DNc
    is_unpatchable_builtin_method = inspect.ismethoddescriptor(
        obj
    ) and not obj.__class__.__name__.startswith("cython")

    return inspect.isroutine(obj) and not is_unpatchable_builtin_method


@fail_quietly("Unable to repatch single module")
def repatch_module(module):
    """Repatch a single module. See docstring for repatch_imported_modules"""

    module_attrs = list(vars(module).items())

    for attr_name, attr in module_attrs:
        try:
            if not is_patchable(attr):
                continue
        except Exception as e:
            if (
                module.__name__.startswith("django")
                and type(e).__name__ == "ImproperlyConfigured"
            ):
                # Django gives ImproperlyConfigured if present in the env but unused.
                # This is not an issue, but can lead to noisy logging.
                continue
            logger.debug(
                "exception occurred while checking whether to patch %s in %s",
                attr_name,
                module.__name__,
                exc_info=e,
            )
            continue
        if type(attr) is functools.partial and (
            orig_patch := patch_manager.get_patch(attr.func)
        ):
            patch = functools.partial(orig_patch, *attr.args, **attr.keywords)
        else:
            patch = patch_manager.get_outer_patch(patch_manager.as_func(attr))

        if patch:
            logger.debug("applying repatch to %s in %s", attr_name, module.__name__)
            patch_manager.patch(module, attr_name, patch)


@fail_quietly("Unable to patch previously imported modules")
def repatch_imported_modules():
    """
    Applies patches to modules that were already imported prior to agent startup

    Here's the problem: our patches don't get applied until after our
    middleware class is initialized. At this point it's likely that most (or
    all) application modules will have already been imported.

    If we patch the function `foo.bar.baz`, and an application module that was
    loaded prior to our patches imports it as `from foo.bar import baz`, then
    our patch will have no effect within that application module. This is
    because the application module has a reference to the *original* function,
    and that reference remains unchanged even after we apply a patch to the
    `foo.bar` module.

    On the other hand, if the application imports it as `from foo import bar`
    and uses it as `bar.baz()`, then our patches will work just fine. In this
    case, the application module has a reference to the *module itself*, which
    is where we apply our patch. This means that when the application calls
    `bar.baz()`, it will be calling the updated (patched) function.

    Incidentally, if the application imports as `from foo.bar import baz`, but
    this module is not loaded until *after* our patches have been applied, our
    patch will be effective. However, we have no control over the order of
    imports in an application.

    This function is designed to remedy the former case in order to make sure
    that our patches are effective regardless of how they are imported or the
    order in which they are imported by the application.

    Prior to calling this function, we make a record of every function that
    gets patched. After all patches are applied, this function iterates
    through all imported modules, which includes all modules that may have been
    imported before our patches were applied. We look for any instances of the
    original functions that need to be patched, and we replace them with the
    patches in those modules.
    """
    for module in get_loaded_modules().values():
        repatch_module(module)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import inspect

NOTIMPLEMENTED_MSG = "This method should be implemented by concrete subclass subclass"


def get_name(obj):
    return f"{obj.__module__}.{obj.__name__}" if inspect.isclass(obj) else obj.__name__



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations


from functools import cached_property
import threading
import contextlib

from contrast_fireball import (
    AssessFinding,
    AssessRoute,
    AssessRouteObservation,
    Browser,
    ObservedRoute,
)

import contrast
from contrast.reporting import teamserver_messages
from contrast.agent import (
    agent_state,
    scope,
    thread_watcher,
    request_state,
)
from contrast.agent.protect import input_analysis
from contrast.agent.request_context import RequestContext
from contrast.agent.assess.rules.response.xss import analyze_xss
from contrast.agent.assess.rules.response.analyze import analyze_response_rules
from contrast.agent.assess.preflight import update_preflight_hashes
from contrast.agent.middlewares import response_wrappers
from contrast.agent.middlewares.response_wrappers.base_response_wrapper import (
    BaseResponseWrapper,
)
from contrast.api.attack import ProtectResponse
from contrast.utils.loggers.logger import (
    setup_basic_agent_logger,
)
from contrast.utils.decorators import fail_loudly, fail_quietly
from contrast.utils import timer
from contrast.utils.monitoring import Profiler, tracer

# initialize a basic logger until config is parsed
logger = setup_basic_agent_logger()


@contextlib.contextmanager
def log_request_start_and_end(request_method, request_path):
    start_time = timer.now_ms()
    logger.debug(
        "Beginning request analysis",
        request_method=request_method,
        request_path=request_path,
    )
    try:
        yield
    finally:
        logger.debug(
            "Ending request analysis",
            request_method=request_method,
            request_path=request_path,
        )
        logger.info(
            "request summary",
            request_path=request_path,
            request_method=request_method,
            elapsed_time_ms=timer.now_ms() - start_time,
            # native thread id is useful for lining up viztracer threads to requests,
            # but it requires a syscall so we don't want it in every log message
            native_thread_id=threading.get_native_id(),
        )


def _log_response_info(response: BaseResponseWrapper):
    if response is None:
        logger.debug("No response info for this request")
        return
    logger.debug(
        "Response summary",
        status_code=response.status_code,
        content_length=response.headers.get("content-length", ""),
    )


class BaseMiddleware:
    """
    BaseMiddleware contains all the initial setup for the framework middlewares

    Requirements:

        1. It's callable
        2. It has call_with_agent
        3. It has call_without_agent

    Pre and post filter calls should not block the flow that this class has.

    Pre -> get_response -> post
    """

    # TODO: PYT-2852 Revisit application name detection
    app_name = ""  # This should be overridden by child classes

    DIAGNOSTIC_ENDPOINT = "/save-contrast-security-config"
    DIAGNOSTIC_ALLOWED_SERVER = "localhost"
    DIAGNOSTIC_ALLOWED_IP = "127.0.0.1"

    @scope.contrast_scope()
    def __init__(self):
        """
        Most agent initialization is now done by the agent state module

        This method calls agent state initialization if it hasn't been done
        already and then loads the settings and reporting client.

        Config scanning still happens here since the behavior is framework-specific.
        """
        # id will be different across processes but also for multiple middlewares
        # within the same process
        self.id = id(self)
        self.settings = None
        self.request_start_time = None

        agent_state.initialize()

        self.settings = agent_state.get_settings()
        if not self.is_agent_enabled():
            return

        self.reporting_client = agent_state.module.reporting_client

    @contextlib.contextmanager
    def request_context_stack(self, request_method: str, request_path: str):
        """
        Create a combined context manager for request processing that includes:
        - request_id_context
        - cProfiler (if enabled)
        - log_request_start_and_end
        - VizTracer (if enabled)

        This consolidates duplicate code across middleware implementations.
        """
        with contextlib.ExitStack() as stack:
            # the request_id context manager must come first!
            stack.enter_context(request_state.request_id_context())
            stack.enter_context(log_request_start_and_end(request_method, request_path))

            # TODO: PYT-3337 - we should issue a deprecation warning if
            # agent.python.enable_profiler is used
            profiler_enabled = self.settings.config.get(
                "agent.python.enable_profiler"
            ) or self.settings.config.get("agent.python.profiler.enable")
            if profiler_enabled:
                stack.enter_context(Profiler(request_path))
            tracer_enabled = self.settings.config.get("agent.python.tracer.enable")
            if tracer_enabled:
                min_duration_ms = self.settings.config.get(
                    "agent.python.tracer.min_duration_ms", 0
                )
                stack.enter_context(tracer(request_path, min_duration_ms))

            yield stack

    def enter_agent_call_contexts(
        self, stack: contextlib.ExitStack, context: RequestContext
    ):
        """
        Enter the pre-request contexts expected before calling the handler with the agent.
        """
        stack.enter_context(contrast.lifespan(context))
        if context.observe_enabled and self.reporting_client is not None:
            context.observability_trace = stack.enter_context(
                self.reporting_client.observability_trace(send_trace=True)
            )
        return stack

    @cached_property
    def name(self):
        raise NotImplementedError("Must implement name")

    def is_agent_enabled(self):
        """
        Agent is considered enabled if the config value for 'enable' is True (or empty,
        defaults to True). Errors during initialization or runtime will set this value
        to False.
        """
        if self.settings is None:
            return False

        return self.settings.is_agent_config_enabled()

    def call_with_agent(self, *args):
        raise NotImplementedError("Must implement call_with_agent")

    def call_without_agent(self, *args):
        """
        The agent does not set context when this function is called so all other patches
        (e.g propagators) that check context shouldn't run.
        """
        raise NotImplementedError("Must implement call_without_agent")

    def should_analyze_request(self, environ) -> RequestContext | None:
        """
        Determine if request should be analyzed based on configured settings.

        While returning different types of objects based on logic is not a good
        pattern, in this case it's an optimization that allows us to create
        the request context obj when we need it.
        """
        path = environ.get("PATH_INFO")

        if not self.is_agent_enabled():
            logger.debug("Will not analyze request: agent disabled.", path=path)
            return None

        # TODO: PYT-3778 As an optimization, we could skip analysis here (return False)
        # if none of the individual agent modes are enabled

        context = RequestContext(
            environ,
            assess_enabled=agent_state.module.assess_enabled,
            exclusions=agent_state.module.exclusions,
            request_data_masker=agent_state.module.request_data_masker,
            event_handlers=agent_state.module.event_handlers,
            observe_enabled=agent_state.module.observe_enabled,
            protect_enabled=agent_state.module.protect_enabled,
        )

        if context.evaluate_exclusions():
            logger.debug(
                "Will not analyze request: request meets exclusions.", path=path
            )
            return None

        from contrast.agent.assess import sampling

        if sampling.meets_criteria(context, agent_state.module.sampling_cfg):
            logger.debug(
                "Will not run Assess analysis on request: request meets sampling.",
                path=path,
            )
            context.assess_enabled = False

        return context

    @fail_loudly("Unable to do handle_ensure")
    def handle_ensure(self, context: RequestContext, request):
        """
        Method that should run for all middlewares AFTER every request is made.
        """
        thread_watcher.ensure_running(agent_state.module)

        if request is not None:
            if context.assess_enabled:
                self._handle_observed_route(context, request)
                update_preflight_hashes(context)
            if context.observe_enabled and (
                (trace := context.observability_trace) is not None
            ):
                http_span_attrs = context.request.get_otel_attributes()
                if context.path_template is not None:
                    http_span_attrs["http.route"] = context.path_template
                if context.response:
                    http_span_attrs.update(
                        response_wrappers.get_otel_attributes(context.response)
                    )
                if context.response_exception is not None:
                    error_type = type(context.response_exception)
                    http_span_attrs["error.type"] = (
                        f"{error_type.__module__}.{error_type.__qualname__}"
                    )
                trace.update(http_span_attrs)

        logger.debug("Sending final messages for reporting.")
        for msg in self.final_ts_messages(context):
            self.reporting_client.add_message(msg)

        self.send_library_observations(context)
        self.send_findings(context)
        self.send_route_observation(context)
        self.send_inventory_components(context)

        _log_response_info(context.response)

        agent_state.set_first_request(False)

    def final_ts_messages(self, context):
        if context is None:
            return []

        messages = []
        if context.attacks:
            messages.append(
                teamserver_messages.ApplicationActivity(
                    attacks=context.attacks,
                    request=context.request,
                    request_data_masker=context.request_data_masker,
                )
            )

        return messages

    def send_library_observations(self, context):
        if library_observations := context.observed_library_usage:
            self.reporting_client.new_library_observations(library_observations)

    def send_findings(self, context):
        if (
            context.assess_enabled
            and context.findings is not None
            and len(context.findings) > 0
        ):
            # Masking can be expensive for large requests, so we only do it if
            # we're sending a message that requires masking, such as findings (here)
            # or application activity (for Protect attack samples).
            if context.request_data_masker:
                context.request_data_masker.mask_sensitive_data(context.request)

            self.reporting_client.new_findings(context.findings, context.request)

    def send_route_observation(self, context):
        # Per the spec, we do not report an observed route if the route signature or URL is empty.
        # Also, we don't report on a subset of error response codes. (PYT-3306)
        if (
            context.assess_enabled
            and context.response is not None
            and self._desired_observation_response_code(context.response.status_code)
            and context.observed_route.signature
            and context.observed_route.url
        ):
            self.reporting_client.new_observed_route(context.observed_route)

    def send_inventory_components(self, context: RequestContext):
        if (
            self.reporting_client is not None
            and context.request is not None
            and context.request.user_agent
        ):
            self.reporting_client.new_inventory_components(
                [Browser(context.request.user_agent)]
            )

    @fail_loudly("Failed to run prefilter.")
    def prefilter(self, context: RequestContext):
        """
        Run all of our prefilter, those that happen before handing execution to the application code, here.
        """
        if context.protect_enabled:
            self.prefilter_protect()

    @fail_quietly("Failed to run prefilter protect.")
    def prefilter_protect(self):
        """
        Prefilter - AKA input analysis - is performed mostly with agent-lib but partly in the agent.

        In this method we call on agent-lib to do input analysis, which can result in:
        1. agent-lib finds an attack in which case we block the request
        2. agent-lib returns input analysis to use for later sink / infilter analysis, in which case we store it here
          in the request context.

        We then call to each rule to determine if they have any special prefilter actions, whether due to not being
        implemented in agent-lib or b/c they need special analysis.
        """
        logger.debug("PROTECT: Running Agent prefilter.")

        input_analysis.analyze_inputs()

        for rule in self.settings.protect_rules.values():
            if rule.is_prefilter():
                rule.prefilter()

    @fail_loudly("Unable to do postfilter")
    def postfilter(self, context):
        """
        For all postfilter enabled rules.
        """
        if context.protect_enabled:
            self.postfilter_protect(context)

        if context.assess_enabled:
            self.response_analysis(context)

    def _process_trigger_handler(self, handler):
        """
        Gather metadata about response handler callback for xss trigger node

        We need to check whether the response handler callback is an instance method or
        not. This affects the way that our policy machinery works, and it also affects
        reporting, so we need to make sure to account for the possibility that handler
        is a method of some class rather than a standalone function.

        This should be called by the `trigger_node` method in child classes.
        """
        module = handler.__module__
        class_name = ""

        if hasattr(handler, "__self__"):
            class_name = handler.__self__.__class__.__name__
            args = (handler.__self__,)
            instance_method = True
        else:
            args = ()
            instance_method = False

        return module, class_name, args, instance_method

    @cached_property
    def trigger_node(self):
        """
        Trigger node property used by assess reflected xss postfilter rule

        This must be overridden by child classes that make use of the reflected
        xss postfilter rule.
        """
        raise NotImplementedError("Children must define trigger_node property")

    @fail_loudly("Unable to do assess response analysis")
    def response_analysis(self, context):
        """
        Run postfilter for any assess rules. Reflected xss rule runs by default.
        May be overridden in child classes.

        If the response content type matches a allowed content type, do not run
        assess xss response analysis. This is because the security team
        considers reflected xss within these content types to be a false positive.
        """
        logger.debug("ASSESS: Running response analysis")

        analyze_xss(context, self.trigger_node)
        analyze_response_rules(context)

    @fail_quietly("Failed to run postfilter protect.")
    def postfilter_protect(self, context):
        logger.debug("PROTECT: Running Agent postfilter.")

        for rule in self.settings.protect_rules.values():
            if rule.is_postfilter():
                rule.postfilter()

    @fail_loudly("Unable to do check_for_blocked")
    def check_for_blocked(self, context: RequestContext):
        """
        Checks for BLOCK events in case SecurityException was caught by app code

        This should be called by each middleware after the view is generated
        but before returning the response (it can be before or after
        postfilter).

        If we make it to this call, it implies that either no SecurityException
        occurred, or if one did occur, it was caught by the application. If we
        find a BLOCK here, it necessarily implies that an attack was detected
        in the application, but the application caught our exception. If the
        application hadn't caught our exception, we never would have made it
        this far because the exception would have already bubbled up to the
        middleware exception handler. So this is really our first and our last
        opportunity to check for this particular edge case.
        """
        for attack in context.attacks:
            if attack.response == ProtectResponse.BLOCKED:
                raise contrast.SecurityException(rule_name=attack.rule_id)

    def _desired_observation_response_code(self, response_code: int) -> bool:
        """
        If we're filtering route observation by response code, determine if the
        one provided is desired or not. We desire any non-error response by
        default, or any response if filtering has been disabled by the user.
        """
        return response_code not in (
            403,
            404,
            405,
            501,
        ) or self.settings.config.get("agent.route_coverage.report_on_error")

    def _handle_observed_route(self, context: RequestContext, request) -> None:
        """
        Perform any necessary actions related to the observed route (for Assess). By the
        time this function is called, framework-specific logic should have identified
        the route signature for this request, if available.
        """
        context.observed_route = ObservedRoute(
            signature=self._get_signature(context, request),
            url=context.request.get_normalized_uri(),
            verb=request.method,
            sources=context.observed_route.sources,
        )
        logger.debug(
            "stored observed route on context",
            observed_route=context.observed_route,
        )
        self._append_route_to_findings(context.observed_route, context.findings)

    def _get_signature(self, context: RequestContext, request) -> str:
        """
        There are a few different strategies we might use to obtain a route signature.
        In order of preference:

        1.  Obtain a signature at some point during the request; this string is set on
            `context.signature`. For supported frameworks, this is accomplished with
            framework-specific patches. One exception to this is aiohttp, where it
            happens directly in the middleware. It is essential that for a particular
            route this signature exactly matches the one found during route discovery.

        2.  If context.signature isn't set, we didn't get any framework-specific
            information about the view function. In this case we use the request's
            normalized URI as the signature. We expect to hit this case in pure WSGI /
            ASGI.
        """
        if context.signature:
            return context.signature

        logger.debug(
            "Did not find a view function signature for the current request. "
            "Falling back on normalized URI."
        )
        return context.request.get_normalized_uri()

    def _append_route_to_findings(
        self, observed_route: ObservedRoute, findings: list[AssessFinding]
    ):
        """
        Append the observed route to any existing findings. We can't necessarily do this
        at finding creation time, because we might not have route info yet.
        """
        if not findings:
            logger.debug("No findings for the current request")
            return

        for finding in findings:
            if not finding.routes:
                logger.debug(
                    "Appending route %s:%s to %s",
                    observed_route.verb,
                    observed_route.url,
                    finding.rule_id,
                )
                finding.routes.append(
                    AssessRoute(
                        count=1,
                        signature=observed_route.signature,
                        observations=[
                            AssessRouteObservation(
                                url=observed_route.url, verb=observed_route.verb or ""
                            )
                        ],
                    )
                )



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations

from contrast.utils.mapping import GlomDict


class NGServerSettings:
    def __init__(self, features: dict | None = None):
        self.features = GlomDict(features or {})

    def common_config(self) -> dict[str, object]:
        """
        Returns the settings in a flattened common configuration format.
        """
        return {
            "server.environment": self.features.get("environment", ""),
            "assess.enable": self.features.get("features.assessment.enabled"),
            "assess.sampling.enable": self.features.get(
                "features.assessment.sampling.enabled", False
            ),
            "assess.sampling.baseline": self.features.get(
                "features.assessment.sampling.baseline"
            ),
            "assess.sampling.request_frequency": self.features.get(
                "features.assessment.sampling.frequency"
            ),
            "assess.sampling.window_ms": (
                window * 1000
                if (window := self.features.get("features.assessment.sampling.window"))
                else None
            ),
            "agent.logger.level": self.features.get("features.logLevel"),
            "agent.logger.path": self.features.get("features.logFile"),
            "agent.security_logger.syslog.enable": self.features.get(
                "features.defend.syslog.syslogEnabled"
            ),
            "agent.security_logger.syslog.protocol": self.features.get(
                "features.defend.syslog.syslogProtocol"
            ),
            # features.defend.syslog.syslogConnectionType is ignored because the protocol
            # includes the connection type.
            "agent.security_logger.syslog.server_host": self.features.get(
                "features.defend.syslog.syslogIpAddress"
            ),
            "agent.security_logger.syslog.port": self.features.get(
                "features.defend.syslog.syslogPortNumber"
            ),
            "agent.security_logger.syslog.facility": self.features.get(
                "features.defend.syslog.syslogFacilityCode"
            ),
            "agent.security_logger.syslog.severity_exploited": self.features.get(
                "features.defend.syslog.syslogSeverityExploited"
            ),
            "agent.security_logger.syslog.severity_blocked": self.features.get(
                "features.defend.syslog.syslogSeverityBlocked"
            ),
            # there is no severity_blocked_perimeter in the ng version of the endpoint.
            "agent.security_logger.syslog.severity_probed": self.features.get(
                "features.defend.syslog.syslogSeverityProbed"
            ),
            "agent.security_logger.syslog.severity_suspicious": self.features.get(
                "features.defend.syslog.syslogSeveritySuspicious"
            ),
            "protect.enable": self.features.get("features.defend.enabled", False),
        }



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from typing import TypedDict

# Unfortunately, TypedDicts do not currently support arbitrary extra keys in addition to
# required keys, so we cannot use one here.
EventDict = dict
"""
Part of a v2 policy definition that contains any metadata required to build event
handler functions. At minimum, has an event `name` key.
"""


class PolicyDefinition(TypedDict):
    """
    v2 policy definition for a group of functions that share an event type. Used for
    literal contrast-defined policy.

    Try to keep this easily JSON-serializable in case we want to support receiving
    policy definitions from external sources in the future.
    """

    module: str
    method_names: list[str]
    event: EventDict


def definitions() -> list[PolicyDefinition]:
    """
    Returns a list of all v2 policy definitions.
    """
    return cmd_exec + file_open + authn + authz + outbound_request


cmd_exec: list[PolicyDefinition] = [
    {
        "module": "os",
        "method_names": ["system"],
        "event": {
            "name": "cmd-exec",
            "cmd": "command",
        },
    },
    {
        "module": "subprocess",
        "method_names": ["Popen.__init__"],
        "event": {
            "name": "cmd-exec",
            "cmd": "executable",
            "args": "args",
            "shell": "shell",
        },
    },
    {
        "module": "os",
        "method_names": ["spawnv", "spawnvp", "spawnve", "spawnvpe"],
        "event": {
            "name": "cmd-exec",
            "args": "args",
            # There could be an argument that we should include "file" as the "cmd".
            # This will often be the same as args[0] though, and we don't want the command
            # string to stutter. Using just "args" is likely to be closer to the user's
            # intent.
        },
    },
]

file_open: list[PolicyDefinition] = [
    {
        "module": "builtins",
        "method_names": ["open"],
        "event": {
            "name": "file-open",
            "file": "file",
            "flags": "mode",
        },
    },
    {
        "module": "os",
        "method_names": ["open"],
        "event": {
            "name": "file-open",
            "file": "path",
            "flags": "flags",
        },
    },
    {
        "module": "dbm.dumb",
        "method_names": ["open"],
        "event": {
            "name": "file-open",
            "file": "file",
            "flags": "flag",
            "dbm": True,
        },
    },
    {
        "module": "dbm.gnu",
        "method_names": ["open"],
        "event": {
            "name": "file-open",
            "file": "filename",
            "flags": "flags",
            "dbm": True,
        },
    },
    {
        "module": "dbm.ndbm",
        "method_names": ["open"],
        "event": {
            "name": "file-open",
            "file": "filename",
            "flags": "flags",
            "dbm": True,
        },
    },
]

authn: list[PolicyDefinition] = [
    {
        "module": "django.contrib.auth",
        "method_names": ["authenticate", "aauthenticate"],
        "event": {
            "name": "django-authn",
            "mechanisms": ["token", "password"],
        },
    },
    {
        "module": "django.contrib.auth",
        "method_names": ["get_user", "aget_user"],
        "event": {
            "name": "django-session-authn",
            "mechanism": "token",
        },
    },
    {
        "module": "starlette.middleware.authentication",
        "method_names": ["AuthenticationMiddleware.__call__"],
        "event": {
            "name": "starlette-authn",
        },
    },
]

authz: list[PolicyDefinition] = [
    {
        "module": "django.contrib.auth.models",
        "method_names": [
            "PermissionsMixin.has_perm",
            "PermissionsMixin.ahas_perm",
            "AnonymousUser.has_perm",
            "AnonymousUser.ahas_perm",
        ],
        "event": {
            "name": "authz",
            "dac_perm": "perm",
        },
    },
    {
        "module": "django.contrib.auth.models",
        "method_names": [
            "PermissionsMixin.has_module_perms",
            "PermissionsMixin.ahas_module_perms",
        ],
        "event": {
            "name": "authz",
            "dac_perm": "app_label",
        },
    },
    {
        "module": "django.contrib.auth.models",
        "method_names": [
            "AnonymousUser.has_module_perms",
            "AnonymousUser.ahas_module_perms",
        ],
        "event": {
            "name": "authz",
            "dac_perm": "module",
        },
    },
    {
        "module": "django.contrib.auth.models",
        "method_names": ["UserManager.with_perm"],
        "event": {
            "name": "authz",
            "dac_perm": "perm",
        },
    },
    {
        "module": "starlette.authentication",
        "method_names": ["has_required_scope"],
        "event": {
            "name": "authz",
            "dac_perms": "scopes",
        },
    },
]

outbound_request: list[PolicyDefinition] = [
    {
        "module": "urllib.request",
        "method_names": ["urlopen"],
        "event": {
            "name": "outbound-request",
            "url": "url",
        },
    },
    {
        "module": "http.client",
        "method_names": ["HTTPConnection.putrequest"],
        "event": {
            "name": "outbound-request-http.client",  # special case for http.client
        },
    },
    {
        "module": "httpx._client",
        "method_names": ["BaseClient.build_request"],
        "event": {
            "name": "outbound-request",
            "url": "url",  # str | httpx.URL
        },
    },
    {
        "module": "aiohttp",
        "method_names": ["ClientSession._request"],
        "event": {
            "name": "outbound-request",
            "url": "str_or_url",  # str | yarl.URL
        },
    },
    {
        "module": "requests.sessions",
        "method_names": [
            "Session.request",
        ],
        "event": {
            "name": "outbound-request",
            "url": "url",
        },
    },
]



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from collections import namedtuple
from html.parser import HTMLParser

from contrast.agent.settings import Settings
from contrast_vendor import structlog as logging
from contrast.utils.string_utils import ensure_string

logger = logging.getLogger("contrast")

Tag = namedtuple("Tag", ["type", "tag", "attrs"])


def analyze_response_rules(context):
    settings = Settings()

    response_rules = settings.enabled_response_rules()
    response = context.response
    if (
        not response_rules
        or not response
        or not response.body
        or not (body := ensure_string(response.body))
    ):
        return

    content_type = response.headers.get("content-type", "")

    status_code = response.status_code
    valid_response_rules = [
        rule for rule in response_rules if rule.is_valid(status_code, content_type)
    ]

    if not valid_response_rules:
        return

    form_tags, meta_tags = get_tags(body, content_type)

    logger.debug("Analyzing response rules", response_rules=valid_response_rules)

    for rule in valid_response_rules:
        violated, properties = rule.is_violated(
            response.headers, body, form_tags, meta_tags
        )
        if violated:
            rule.build_and_append_finding(properties, context)


class BodyParser(HTMLParser):
    def __init__(self):
        super().__init__()
        self.form_tags = []
        self.meta_tags = []

    def handle_starttag(self, tag, attrs):
        if tag == "form":
            self.form_tags.append(
                Tag(type="form", tag=self.get_starttag_text(), attrs=attrs)
            )
        if tag == "meta":
            self.meta_tags.append(
                Tag(type="meta", tag=self.get_starttag_text(), attrs=attrs)
            )

    def error(self, message: str):
        """
        error is a method that should be overridden by a subclass to handle
        HTMLParser errors. In py3.10, it is no longer used and the parser
        will raise AssertionErrors instead. This method unifies the behavior
        across versions.
        """
        raise AssertionError(message)


def get_tags(body, content_type: str):
    parser = BodyParser()
    try:
        parser.feed(body)
        parser.close()
    except AssertionError as exc:
        # Invalid HTML throws assertion errors. HTMLParser doesn't expose an API
        # to disable these assertions.
        # We don't validate that the body is HTML, and the error can be ignored
        # if the body isn't HTML. If the body is HTML or we partially parsed
        # tags, throw the exception so that we can investigate the issue.
        if "html" in content_type:
            raise RuntimeError(
                f"Failed to parse response body {content_type=}"
            ) from exc
        if parser.form_tags or parser.meta_tags:
            raise RuntimeError(
                f"Partially parsed response body {parser.form_tags=} {parser.meta_tags=}"
            ) from exc

    return parser.form_tags, parser.meta_tags



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.response.base_response_rule import BaseResponseRule
from contrast.agent.assess.rules import build_finding, send_finding


class BaseHeaderOnlyRule(BaseResponseRule):
    @property
    def name(self):
        raise NotImplementedError

    @property
    def header_key(self):
        raise NotImplementedError

    @property
    def good_values(self):
        raise NotImplementedError

    def is_violated(self, headers, *args):
        header_violated, header_properties = self.is_header_violated(headers)
        if not header_violated:
            return False, {}

        return True, header_properties

    def is_header_violated(self, headers):
        value = headers.get(self.header_key)
        if not value:
            # Finding gets no properties if header is missing
            return True, {}

        if value not in self.good_values:
            properties = dict(
                type="Header",
                name=self.header_key,
                value=value,
            )
            return True, properties

        return False, None

    def create_finding(self, properties):
        """
        header-only rules report data=str, where str is the value
        the header is set to. This is unlike other response rules that
        report data=[]
        """
        if not properties:
            data = dict(data="")
        else:
            data = dict(data=f"{properties.get('value')}")
        return build_finding(self, data)

    def build_and_append_finding(self, properties, context):
        finding = self.create_finding(properties)
        send_finding(finding, context)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from collections import OrderedDict
from collections.abc import Mapping

from contrast.agent.protect.rule.base_rule import BaseRule
from contrast.agent.protect.rule.cmdi_rule import CmdInjection
from contrast.agent.protect.rule.deserialization_rule import Deserialization
from contrast.agent.protect.rule.http_method_tampering import MethodTampering
from contrast.agent.protect.rule.malformed_header import MalformedHeader
from contrast.agent.protect.rule.nosqli_rule import NoSqlInjection
from contrast.agent.protect.rule.path_traversal_rule import PathTraversal
from contrast.agent.protect.rule.sqli_rule import SqlInjection
from contrast.agent.protect.rule.ssrf_rule import Ssrf
from contrast.agent.protect.rule.unsafe_file_upload_rule import UnsafeFileUpload
from contrast.agent.protect.rule.xss_rule import Xss
from contrast.agent.protect.rule.xxe_rule import Xxe
from contrast.agent.protect.rule.bot_blocker_rule import BotBlocker


def build_protect_rules() -> Mapping[str, BaseRule]:
    """
    Build a dict with rules with prefilter rules first.
    We want prefilter rules first so they get evaluated / trigger first.

    :return: an ordered dict of protect rules
    """
    rules = OrderedDict(
        {
            UnsafeFileUpload.RULE_NAME: UnsafeFileUpload(),
            BotBlocker.RULE_NAME: BotBlocker(),
            CmdInjection.RULE_NAME: CmdInjection(),
            Deserialization.RULE_NAME: Deserialization(),
            # Turned off until TS can handle rule information
            MalformedHeader.RULE_NAME: MalformedHeader(),
            MethodTampering.RULE_NAME: MethodTampering(),
            NoSqlInjection.RULE_NAME: NoSqlInjection(),
            PathTraversal.RULE_NAME: PathTraversal(),
            SqlInjection.RULE_NAME: SqlInjection(),
            Ssrf.RULE_NAME: Ssrf(),
            Xss.RULE_NAME: Xss(),
            Xxe.RULE_NAME: Xxe(),
        }
    )

    return rules



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys
from contrast.agent.policy import patch_manager
from contrast.patches.databases import dbapi2
from contrast.utils.patch_utils import (
    register_module_patcher,
    unregister_module_patcher,
)

PYMYSQL = "pymysql"
VENDOR = "MySQL"


def instrument_pymysql(pymysql):
    dbapi2.instrument_adapter(
        pymysql, VENDOR, dbapi2.Dbapi2Patcher, extra_cursors=[pymysql.cursors.Cursor]
    )


def register_patches():
    register_module_patcher(instrument_pymysql, PYMYSQL)


def reverse_patches():
    unregister_module_patcher(PYMYSQL)
    if pymysql := sys.modules.get(PYMYSQL):
        patch_manager.reverse_patches_by_owner(pymysql)
        patch_manager.reverse_patches_by_owner(pymysql.connections.Connection)
        patch_manager.reverse_patches_by_owner(pymysql.cursors.Cursor)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import math

from contrast.agent.assess.policy.propagators.base_propagator import BasePropagator
from contrast.agent.assess.utils import copy_from


class CenterPropagator(BasePropagator):
    def _propagate(self):
        start_index = math.ceil((len(self.target) - len(self.first_source)) / 2)
        copy_from(self.target, self.first_source, start_index, self.node.untags)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations
from collections.abc import Mapping
import os
from typing import Callable

from contrast_fireball import OtelAttributes, SpanType

from contrast.agent.policy.handlers import EventDict


def observe_span_attrs_builder(
    event_dict: EventDict,
) -> tuple[
    SpanType,
    Callable[[Mapping[str, object], object], OtelAttributes],
    Callable[[Mapping[str, object]], str | None],
]:
    file_param_location = event_dict.get("file", "")
    flags_param_location = event_dict.get("flags", "")
    is_dbm = event_dict.get("dbm", False)

    if not file_param_location or not flags_param_location:
        raise ValueError(
            "Event must specify both 'file' and 'flags' parameter locations."
        )

    def skip_check(args: Mapping[str, object]) -> str | None:
        file_path = args.get(file_param_location, "")
        if not file_path:
            return "empty file path"
        if isinstance(file_path, int):
            return "path is a file descriptor"
        return None

    OS_FLAG_MASK = os.O_RDONLY | os.O_WRONLY | os.O_RDWR

    def observe_span_attrs(args: Mapping[str, object], result: object):
        file_path = args.get(file_param_location, "")
        fs_path = os.fspath(file_path)
        abs_file_path = os.path.abspath(fs_path)
        if isinstance(abs_file_path, bytes):
            abs_file_path = abs_file_path.decode("utf-8", errors="replace")

        flags = args.get(flags_param_location, "")
        str_flags = ""
        if isinstance(flags, int):
            masked_flags = flags & OS_FLAG_MASK
            if masked_flags == os.O_RDONLY:
                str_flags = "o_rdonly"
            elif masked_flags == os.O_WRONLY:
                str_flags = "o_wronly"
            elif masked_flags == os.O_RDWR:
                str_flags = "o_rdwr"
        if isinstance(flags, str):
            if "+" in flags or flags in ("c", "n"):
                str_flags = "o_rdwr"
            elif "r" in flags:
                str_flags = "o_rdonly"
            elif "w" in flags or "a" in flags or "x" in flags:
                str_flags = "o_wronly"
                # special case: https://docs.python.org/3/library/dbm.html#dbm.open
                if is_dbm and flags == "w":
                    str_flags = "o_rdwr"
        if not str_flags:
            raise ValueError(f"Unsupported file open mode: {flags}.")

        return {"file.open.path": abs_file_path, "file.open.flags": str_flags}

    return SpanType.FileOpenCreate, observe_span_attrs, skip_check



# SPDX-License-Identifier: MIT OR Apache-2.0
# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the MIT License.  See the LICENSE file in the root of this
# repository for complete details.

"""
Logger wrapper and helper class.
"""

from __future__ import annotations

from typing import Any, Iterable, Mapping, Sequence

from contrast_vendor.structlog.exceptions import DropEvent

from .typing import BindableLogger, Context, Processor, WrappedLogger


class BoundLoggerBase:
    """
    Immutable context carrier.

    Doesn't do any actual logging; examples for useful subclasses are:

    - the generic `BoundLogger` that can wrap anything,
    - `structlog.stdlib.BoundLogger`.
    - `structlog.twisted.BoundLogger`,

    See also `custom-wrappers`.
    """

    _logger: WrappedLogger
    """
    Wrapped logger.

    .. note::

        Despite underscore available **read-only** to custom wrapper classes.

        See also `custom-wrappers`.
    """

    def __init__(
        self,
        logger: WrappedLogger,
        processors: Iterable[Processor],
        context: Context,
    ):
        self._logger = logger
        self._processors = processors
        self._context = context

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}(context={self._context!r}, processors={self._processors!r})>"

    def __eq__(self, other: object) -> bool:
        try:
            return self._context == other._context  # type: ignore[attr-defined]
        except AttributeError:
            return False

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def bind(self, **new_values: Any) -> BoundLoggerBase:
        """
        Return a new logger with *new_values* added to the existing ones.
        """
        return self.__class__(
            self._logger,
            self._processors,
            self._context.__class__(self._context, **new_values),
        )

    def unbind(self, *keys: str) -> BoundLoggerBase:
        """
        Return a new logger with *keys* removed from the context.

        Raises:
            KeyError: If the key is not part of the context.
        """
        bl = self.bind()
        for key in keys:
            del bl._context[key]

        return bl

    def try_unbind(self, *keys: str) -> BoundLoggerBase:
        """
        Like :meth:`unbind`, but best effort: missing keys are ignored.

        .. versionadded:: 18.2.0
        """
        bl = self.bind()
        for key in keys:
            bl._context.pop(key, None)

        return bl

    def new(self, **new_values: Any) -> BoundLoggerBase:
        """
        Clear context and binds *new_values* using `bind`.

        Only necessary with dict implementations that keep global state like
        those wrapped by `structlog.threadlocal.wrap_dict` when threads
        are re-used.
        """
        self._context.clear()

        return self.bind(**new_values)

    # Helper methods for sub-classing concrete BoundLoggers.

    def _process_event(
        self, method_name: str, event: str | None, event_kw: dict[str, Any]
    ) -> tuple[Sequence[Any], Mapping[str, Any]]:
        """
        Combines creates an ``event_dict`` and runs the chain.

        Call it to combine your *event* and *context* into an event_dict and
        process using the processor chain.

        Args:
            method_name:
                The name of the logger method.  Is passed into the processors.

            event:
                The event -- usually the first positional argument to a logger.

            event_kw:
                Additional event keywords.  For example if someone calls
                ``log.info("foo", bar=42)``, *event* would to be ``"foo"`` and
                *event_kw* ``{"bar": 42}``.

        Raises:
            structlog.DropEvent: if log entry should be dropped.

            ValueError:
                if the final processor doesn't return a str, bytes, bytearray,
                tuple, or a dict.

        Returns:
             `tuple` of ``(*args, **kw)``

        .. note::
            Despite underscore available to custom wrapper classes.

            See also `custom-wrappers`.

        .. versionchanged:: 14.0.0
            Allow final processor to return a `dict`.
        .. versionchanged:: 20.2.0
            Allow final processor to return `bytes`.
        .. versionchanged:: 21.2.0
            Allow final processor to return a `bytearray`.
        """
        # We're typing it as Any, because processors can return more than an
        # EventDict.
        event_dict: Any = self._context.copy()
        event_dict.update(**event_kw)

        if event is not None:
            event_dict["event"] = event
        for proc in self._processors:
            event_dict = proc(self._logger, method_name, event_dict)

        if isinstance(event_dict, (str, bytes, bytearray)):
            return (event_dict,), {}

        if isinstance(event_dict, tuple):
            # In this case we assume that the last processor returned a tuple
            # of ``(args, kwargs)`` and pass it right through.
            return event_dict

        if isinstance(event_dict, dict):
            return (), event_dict

        msg = (
            "Last processor didn't return an appropriate value.  "
            "Valid return values are a dict, a tuple of (args, kwargs), bytes, or a str."
        )
        raise ValueError(msg)

    def _proxy_to_logger(
        self, method_name: str, event: str | None = None, **event_kw: Any
    ) -> Any:
        """
        Run processor chain on event & call *method_name* on wrapped logger.

        DRY convenience method that runs :func:`_process_event`, takes care of
        handling :exc:`structlog.DropEvent`, and finally calls *method_name* on
        :attr:`_logger` with the result.

        Args:
            method_name:
                The name of the method that's going to get called.  Technically
                it should be identical to the method the user called because it
                also get passed into processors.

            event:
                The event -- usually the first positional argument to a logger.

            event_kw:
                Additional event keywords.  For example if someone calls
                ``log.info("foo", bar=42)``, *event* would to be ``"foo"`` and
                *event_kw* ``{"bar": 42}``.

        .. note::
            Despite underscore available to custom wrapper classes.

            See also `custom-wrappers`.
        """
        try:
            args, kw = self._process_event(method_name, event, event_kw)
            return getattr(self._logger, method_name)(*args, **kw)
        except DropEvent:
            return None


def get_context(bound_logger: BindableLogger) -> Context:
    """
    Return *bound_logger*'s context.

    The type of *bound_logger* and the type returned depend on your
    configuration.

    Args:
        bound_logger: The bound logger whose context you want.

    Returns:
        The *actual* context from *bound_logger*. It is *not* copied first.

    .. versionadded:: 20.2.0
    """
    # This probably will get more complicated in the future.
    return bound_logger._context



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

from contrast.utils.decorators import fail_quietly
from contrast.utils.timer import now_ms
from dataclasses import dataclass
from contrast.configuration import AgentConfig
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")

REQUESTS = dict()
LOGGED_SAMPLING_MESSAGE = False
SAMPLER_ENABLED_CONFIG_NAME = "assess.sampling.enable"


@dataclass
class SamplerConfig:
    assess_enabled: bool
    sampler_enabled: bool
    window_ms: int
    baseline: int
    request_freq: int


DEFAULT_ASSESS_IN_PRODUCTION_SAMPLER_CONFIG = SamplerConfig(
    assess_enabled=True,
    sampler_enabled=True,
    window_ms=3600000,  # 1 hour
    baseline=20,
    request_freq=2147483647,  # In most languages this is the maximum signed 4 byte integer (ints are unbound in python)
)


def _set_assess_in_prod_defaults(agent_config: AgentConfig) -> SamplerConfig:
    for opt_name, default_value in {
        SAMPLER_ENABLED_CONFIG_NAME: DEFAULT_ASSESS_IN_PRODUCTION_SAMPLER_CONFIG.sampler_enabled,
        "assess.sampling.window_ms": DEFAULT_ASSESS_IN_PRODUCTION_SAMPLER_CONFIG.window_ms,
        "assess.sampling.baseline": DEFAULT_ASSESS_IN_PRODUCTION_SAMPLER_CONFIG.baseline,
        "assess.sampling.request_frequency": DEFAULT_ASSESS_IN_PRODUCTION_SAMPLER_CONFIG.request_freq,
    }.items():
        agent_config.get_option(opt_name).ui_value = default_value

    return DEFAULT_ASSESS_IN_PRODUCTION_SAMPLER_CONFIG


def update_sampler_config(agent_config: AgentConfig):
    from contrast.agent import agent_state

    assess_enabled = agent_config.get("assess.enable")
    window = agent_config.get("assess.sampling.window_ms")
    baseline = agent_config.get("assess.sampling.baseline")
    req_freq = agent_config.get("assess.sampling.request_frequency")
    sampler_enabled_value = agent_config.get(SAMPLER_ENABLED_CONFIG_NAME)
    sampler_enabled_opt = agent_config.get_option(SAMPLER_ENABLED_CONFIG_NAME)
    server_environ = agent_state.get_server_environment()

    if (
        server_environ == "PRODUCTION"
        and sampler_enabled_opt.is_definitely_static() is False
        and sampler_enabled_value is False
    ):
        # We only want to apply defaults in production if the sampler was not
        # previously configured (i.e. sampler_enabled_value is False)
        cfg = _set_assess_in_prod_defaults(agent_config)
    else:
        cfg = SamplerConfig(
            assess_enabled,
            sampler_enabled_value,
            window,
            baseline,
            req_freq,
        )

    agent_state.module.sampling_cfg = cfg


def _enabled(config: SamplerConfig):
    global LOGGED_SAMPLING_MESSAGE

    sampling_enabled = config.assess_enabled and config.sampler_enabled

    if not LOGGED_SAMPLING_MESSAGE and sampling_enabled:
        logger.info(
            "Contrast assess.sampling is enabled in your configuration. "
            "Not all requests will be analyzed."
        )
        LOGGED_SAMPLING_MESSAGE = True

    return sampling_enabled


# If we can't determine sampling, it's better to analyze the request.
@fail_quietly("Unable to determine sampling", False)
def meets_criteria(context, config: SamplerConfig) -> bool:
    """
    If a request meets criteria for sampling, agent should not analyze request.

    Criteria is met if request happens inside a sampling time window and request
    count minus baseline doesn't meet frequency setting.
    """
    if not _enabled(config):
        return False

    if context.hash is None:
        # If we cannot determine the uniqueness of this request, it's better
        # to analyze the request
        return False

    if context.hash in REQUESTS:
        history = REQUESTS[context.hash]
    else:
        history = REQUESTS[context.hash] = RequestHistory()

    history.hit()

    # if sampling window has been exceeded, reset it.
    if history.elapsed() >= config.window_ms:
        history.reset()
        return False

    # once hits exceed baseline setting, limit analysis based on frequency setting.
    baseline = config.baseline
    frequency = config.request_freq

    return not (history.hits <= baseline or (history.hits - baseline) % frequency == 0)


class RequestHistory:
    def __init__(self):
        self._start = now_ms()
        self._hit = 0

    @property
    def hits(self):
        return self._hit

    def elapsed(self):
        return now_ms() - self._start

    def hit(self):
        """Increment attempted requests, not analyzed requests"""
        self._hit += 1

    def reset(self):
        self._start = now_ms()
        self._hit = 1

    def __repr__(self):
        return f"Started at: {self._start} - Hit : {self._hit}"



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

"""
This module contains patches to support chaining Contrast instrumentation
with other tools, such as opentelemetry and DataDog's runner commands.
"""

from functools import partial
import os
from contrast.agent.scope import contrast_scope
from contrast.utils.patch_utils import (
    build_and_apply_patch,
    register_module_patcher,
    wrap_and_watermark,
)


def register_patches():
    register_module_patcher(
        partial(
            build_and_apply_patch, attr_name="execl", patch_builder=build_os_execl_hook
        ),
        "os",
    )


def build_os_execl_hook(original_func, policy_node):
    loader_path = os.path.join(os.path.dirname(os.path.dirname(__file__)), "loader")

    @contrast_scope()
    def hook(wrapped, instance, args, kwargs):
        if (path := os.environ.get("PYTHONPATH")) and not path.startswith(loader_path):
            os.environ["PYTHONPATH"] = loader_path + os.path.pathsep + path

        return wrapped(*args, **kwargs)

    return wrap_and_watermark(original_func, hook)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
Core architecture for policy v2.

"Policy" is in-agent data that describes how we want to instrument particular functions.
Policy has undergone several iterations over the years, but "v2" is a major refactor
that is a more fundamental change to the way we define and apply policy-based
instrumentation.

A single patch is applied to each function instrumented with policy v2, regardless of
which agent modes are currently enabled. At call-time, this patch retrieves and calls a
series of handler functions specific to the called function. Handler functions are
generated from policy definitions, and can be regenerated after startup to modify the
agent's behavior - for example, if a mode becomes enabled/disabled, handlers can be
easily added/removed.
"""

from __future__ import annotations

from dataclasses import dataclass
from functools import partial
import inspect
from typing import Callable


import contrast
from contrast.agent import scope
from contrast.agent.policy.handlers import (
    EventDict,
    EventHandler,
    EventHandlerBuilder,
    authn,
    authz,
    cmd_exec,
    file_open,
    outbound_request,
    observe_handler_builder,
)
from contrast.agent.request_context import RequestContext
from contrast.policy_v2 import PolicyDefinition
from contrast.utils.decorators import log_and_report_exception
from contrast.utils.patch_utils import add_watermark
from contrast_vendor import wrapt


NO_RESULT: object = object()
"""
Sentinel used by event handlers. Indicates that the original function did not return a
value (most likely, it raised an exception instead).
"""


EVENT_HANDLER_BUILDERS: dict[str, dict[str, EventHandlerBuilder]] = {
    "cmd-exec": {
        "observe": partial(
            observe_handler_builder, cmd_exec.observe_span_attrs_builder
        ),
    },
    "file-open": {
        "observe": partial(
            observe_handler_builder, file_open.observe_span_attrs_builder
        ),
    },
    "django-authn": {
        "observe": partial(
            observe_handler_builder, authn.django_authn_span_attrs_builder
        ),
    },
    "django-session-authn": {
        "observe": partial(
            observe_handler_builder, authn.django_session_authn_span_attrs_builder
        ),
    },
    "starlette-authn": {
        "observe": partial(
            observe_handler_builder, authn.starlette_authn_span_attrs_builder
        ),
    },
    "authz": {
        "observe": partial(observe_handler_builder, authz.authz_span_attrs_builder),
    },
    "outbound-request": {
        "observe": partial(
            observe_handler_builder, outbound_request.observe_span_attrs_builder
        )
    },
    "outbound-request-http.client": {
        "observe": partial(
            observe_handler_builder, outbound_request.http_client_attrs_builder
        )
    },
}
"""
event name -> {mode -> builder fn}

Central storage for v2 policy builder functions
"""

_policy_v2: dict[str, EventDict] = {}
"""
full function name -> event dict

Central storage for v2 policy definitions at runtime.
"""


@dataclass(frozen=True)
class Location:
    """
    Represents a location of a function in the format 'module.method_name' or
    'module.class_name.method_name'. This is used to uniquely identify a function.
    """

    module: str
    method_name: str
    class_name: str | None = None

    @classmethod
    def from_string(cls, name: str) -> Location:
        parts = name.rsplit(".", maxsplit=2)
        if len(parts) == 2:
            return cls(module=parts[0], method_name=parts[1])
        elif len(parts) == 3:
            # This is a loose check for classname. It assumes we aren't dealing with
            # modules that have names starting with uppercase letters. Our existing policy
            # doesn't have such modules, but we should be careful if we accept policy from
            # users.
            if parts[1][0].isupper():
                return cls(
                    module=parts[0],
                    class_name=parts[1],
                    method_name=parts[2],
                )
            else:
                return cls(
                    module=".".join(parts[:2]), class_name=None, method_name=parts[2]
                )
        else:
            raise ValueError(
                f"Invalid location name '{name}'. Must be in the format 'module.method_name' or 'module.class_name.method_name'."
            )


def register_policy_definitions(definitions: list[PolicyDefinition]) -> None:
    """
    Add the given policy definition to centralized storage for all v2 policy definitions
    """
    new_keys = [
        f"{d['module']}.{method_name}"
        for d in definitions
        for method_name in d["method_names"]
    ]
    internal_duplicates = {k for k in new_keys if new_keys.count(k) > 1}
    if internal_duplicates:
        raise RuntimeError(f"Duplicate policy definitions: {internal_duplicates}")

    new_definitions = {
        f"{d['module']}.{method_name}": d["event"]
        for d in definitions
        for method_name in d["method_names"]
    }
    duplicates = set(_policy_v2.keys()).intersection(new_definitions.keys())
    if duplicates:
        raise RuntimeError(f"Duplicate policy definitions: {duplicates}")

    _policy_v2.update(new_definitions)


def get_policy_locations() -> set[Location]:
    return {Location.from_string(location) for location in _policy_v2}


def generate_policy_event_handlers(
    *,
    assess: bool,
    observe: bool,
    protect: bool,
) -> dict[str, list[EventHandler]]:
    """
    Iterate over all registered policy definitions and (re)generate event handlers.
    """
    # NOTE: we may want to cache builder invocations in the future if performance is bad
    event_handlers = {}
    for location_name, event_dict in _policy_v2.items():
        handler_builders = EVENT_HANDLER_BUILDERS[event_dict["name"]]
        handlers = []
        if observe and (build_observe_handler := handler_builders.get("observe")):
            handlers.append(build_observe_handler(event_dict))
        if assess and (build_assess_handler := handler_builders.get("assess")):
            handlers.append(build_assess_handler(event_dict))
        if protect and (build_protect_handler := handler_builders.get("protect")):
            handlers.append(build_protect_handler(event_dict))
        event_handlers[location_name] = handlers

    return event_handlers


def get_event_handlers(
    location_name: str,
) -> tuple[list[EventHandler], RequestContext | None]:
    """
    Gets all current event handlers for a function. Performs the lookup on the event
    handlers stored on the current request context if one is available, otherwise uses
    agent state.

    To avoid duplicate context lookups in the future, also returns the request context.
    """
    if (context := contrast.REQUEST_CONTEXT.get()) is None:
        from contrast.agent import agent_state

        return agent_state.module.event_handlers.get(location_name, []), None
    return context.event_handlers.get(location_name, []), context


def build_generic_contrast_wrapper(original_func, module_name: str | None = None):
    module_name = module_name or original_func.__module__
    location_name = f"{module_name}.{original_func.__qualname__}"
    assert "contrast" not in location_name, (
        f"Attempting to patch Contrast code: {location_name}"
    )
    bind_args = event_arguments_binder(original_func)

    @wrapt.function_wrapper
    @add_watermark
    def generic_contrast_wrapper(wrapped, instance, args, kwargs):
        """
        Generic wrapper for any function instrumented with v2 policy. This is the top-
        level wrapper that is the single entrypoint for all contrast instrumenation.

        The wrapper looks up relevant event handlers and calls them in order before
        calling the original function. Event handlers are then called again in reverse
        order after the original function call.
        """
        if scope.in_contrast_scope():
            return wrapped(*args, **kwargs)

        with scope.contrast_scope():
            # In the future we need to consider error handling more carefully here.
            # Exceptions raised by our machinery should not affect the original function
            # call. Beware that `@fail_quietly` may not work as expected with generator
            # functions.
            result = NO_RESULT
            bound_args = bind_args(instance, args, kwargs)

            post = []
            event_handlers, context = get_event_handlers(location_name)
            for handler in event_handlers:
                try:
                    gen = handler(context, bound_args.arguments)
                    next(gen)
                except StopIteration:  # noqa: PERF203
                    assert False, "Invalid event handler - did not yield"  # noqa: B011 PT015
                except contrast.SecurityException:
                    raise
                except Exception as e:
                    # If an exception is raised in the event handler, we report it and
                    # continue with the next handler.
                    log_and_report_exception(
                        log_message="Exception in pre event handler",
                        error=e,
                        original_func=original_func,
                        args=args,
                        kwargs=kwargs,
                        log_level="error",
                    )
                else:
                    post.append(gen)

            try:
                with scope.pop_contrast_scope():
                    result = wrapped(*args, **kwargs)
            finally:
                for gen in reversed(post):
                    try:
                        gen.send(result)
                    except StopIteration:  # noqa: PERF203
                        pass
                    except contrast.SecurityException:
                        raise
                    except Exception as e:
                        # If an exception is raised in the event handler, we report it and
                        # continue with the next handler.
                        log_and_report_exception(
                            log_message="Exception in post event handler",
                            error=e,
                            original_func=original_func,
                            args=args,
                            kwargs=kwargs,
                            log_level="error",
                        )
                    else:
                        assert False, "Invalid event handler - more than one yield"  # noqa: B011 PT015

            assert result is not NO_RESULT
            return result

    return generic_contrast_wrapper(original_func)


def event_arguments_binder(func: Callable):
    """
    Returns a function that binds the arguments for func.

    This is used to bind the arguments for the event handlers, so that arguments
    can be retrieved by their names regardless of their order or how they are passed
    in the function call.
    """
    sig = inspect.signature(func)
    if inspect.ismethod(func):
        sig = inspect.signature(func.__func__)

    def _bind(instance, args, kwargs) -> inspect.BoundArguments:
        bound_args = (
            sig.bind(instance, *args, **kwargs)
            if instance is not None
            else sig.bind(*args, **kwargs)
        )
        bound_args.apply_defaults()
        return bound_args

    return _bind



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.dataflow_rule import DataflowRule
from contrast.agent.policy.registry import register_trigger_rule


register_trigger_rule(
    DataflowRule.from_nodes(
        "reflected-xss",
        [],
        disallowed_tags=[
            "CUSTOM_ENCODED_REFLECTED_XSS",
            "CUSTOM_ENCODED",
            "CUSTOM_VALIDATED_REFLECTED_XSS",
            "CUSTOM_VALIDATED",
            "LIMITED_CHARS",
            "BASE64_ENCODED",
            "CSS_ENCODED",
            "CSV_ENCODED",
            "HTML_ENCODED",
            "JAVASCRIPT_ENCODED",
            "JAVA_ENCODED",
            "LDAP_ENCODED",
            "OS_ENCODED",
            "SQL_ENCODED",
            "URL_ENCODED",
            "VBSCRIPT_ENCODED",
            "XML_ENCODED",
            "XPATH_ENCODED",
            "XSS_ENCODED",
        ],
        required_tags=["CROSS_SITE"],
    )
)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations
from collections.abc import Mapping
from typing import Callable

from contrast_fireball import OtelAttributes, SpanType

from contrast.agent.policy.handlers import EventDict


def authz_span_attrs_builder(
    event_dict: EventDict,
) -> tuple[SpanType, Callable[[Mapping[str, object], object], OtelAttributes], None]:
    if dac_perm_param_location := str(event_dict.get("dac_perm", "")):

        def observe_span_attrs(args: Mapping[str, object], result: object):
            attrs = {"contrast.authorization.mechanism": "dac"}
            if perm := str(args.get(dac_perm_param_location, "")):
                # TODO: PYT-3959 - move to list of permissions
                attrs["contrast.authorization.dac.permission"] = perm.lower()
            return attrs
    elif dac_perms_param_location := str(event_dict.get("dac_perms", "")):

        def observe_span_attrs(args: Mapping[str, object], result: object):
            attrs = {"contrast.authorization.mechanism": "dac"}
            if perms := args.get(dac_perms_param_location):
                perms = [str(p).lower() for p in perms]
                attrs["contrast.authorization.dac.permissions"] = perms
            return attrs
    else:
        raise ValueError("Event must specify dac_perm or dac_perms")

    return SpanType.AuthorizationRequest, observe_span_attrs, None



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from .agent_connection import AgentConnection
from .agent_startup import AgentStartup
from .application_activity import ApplicationActivity
from .application_inventory import ApplicationInventory
from .application_update import ApplicationUpdate
from .effective_config import EffectiveConfig
from .heartbeat import Heartbeat
from .library_usage import LibraryUsage
from .observed_route import ObservedRoute
from .preflight import Preflight
from .server_activity import ServerActivity
from .base_ts_message import BaseTsMessage, BaseTsAppMessage, BaseTsServerMessage

__all__ = [
    "AgentConnection",
    "AgentStartup",
    "ApplicationActivity",
    "ApplicationInventory",
    "ApplicationUpdate",
    "EffectiveConfig",
    "Heartbeat",
    "LibraryUsage",
    "ObservedRoute",
    "Preflight",
    "ServerActivity",
    "BaseTsMessage",
    "BaseTsAppMessage",
    "BaseTsServerMessage",
]



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.policy.registry import register_propagation_nodes
from contrast.agent.policy.utils import CompositeNode


framework_propagators = [
    {
        "module": "starlette.staticfiles",
        "class_name": "StaticFiles",
        "method_name": "lookup_path",
        "source": "ARG_0,KWARG:path",
        "target": "RETURN",
        "action": "STARLETTE_LOOKUP_PATH",
        "tags": ["SAFE_PATH"],
    },
    {
        "module": "bottle",
        "method_name": "html_escape",
        "source": "ARG_0,KWARG:string",
        "target": "RETURN",
        "action": "SPLAT",
        "tags": ["HTML_ENCODED"],
        "untags": ["HTML_DECODED"],
    },
    {
        "module": "django.utils.html",
        "method_name": "escape",
        "source": "ARG_0,KWARG:text",
        "target": "RETURN",
        "action": "SPLAT",
        "tags": ["HTML_ENCODED"],
        "untags": ["HTML_DECODED"],
    },
    {
        # Special deadzoned patch implemented outside of policy
        "module": "django.utils.safestring",
        "method_name": "mark_safe",
        "source": "ARG_0,KWARG:s",
        "target": "RETURN",
        "action": "KEEP",
        "policy_patch": False,
    },
    {
        "module": "django.utils._os",
        "method_name": "safe_join",
        "source": "ALL_ARGS",
        "target": "RETURN",
        "action": "SAFE_JOIN_DJANGO",
        "tags": ["SAFE_PATH"],
    },
    {
        "module": "django.core.files.storage",  # for django<4.2
        "class_name": "Storage",
        "method_name": "get_available_name",
        "source": "ARG_0,KWARG:name",
        "target": "RETURN",
        "action": "SPLAT",
        "tags": ["SAFE_PATH"],
    },
    {
        "module": "django.core.files.storage.base",  # for django>=4.2
        "class_name": "Storage",
        "method_name": "get_available_name",
        "source": "ARG_0,KWARG:name",
        "target": "RETURN",
        "action": "SPLAT",
        "tags": ["SAFE_PATH"],
    },
    {
        "module": "django.db.models.base",
        "class_name": "Model",
        "method_name": "__init__",
        "source": "ALL_KWARGS",
        "target": "OBJ",
        "action": "DB_WRITE",
        "tags": ["DATABASE_WRITE"],
    },
    {
        # This is a deadzone property patch. See drf_patches.py.
        # source, target, and action here have no real meaning
        "module": "rest_framework.response",
        "class_name": "Response",
        "method_name": "rendered_content",
        "source": "ARG_0",
        "target": "RETURN",
        "action": "deadzone (placeholder)",
        "policy_patch": False,
    },
    {
        "module": "html",
        "method_name": "escape",
        "source": "ARG_0,KWARG:s",
        "target": "RETURN",
        "action": "SPLAT",
        "tags": ["HTML_ENCODED"],
        "untags": ["HTML_DECODED"],
    },
    {
        # This function was removed in Quart 0.18. There is no sanitizer for python object to json string conversion
        # https://github.com/pallets/quart/compare/0.17.0...0.18.3#diff-8544903fbe8454006675de99682a15abd67c106228ba1864f19b9b379ded9acaL73
        "module": "quart.json",
        "method_name": "htmlsafe_dumps",
        "source": "ARG_0,KWARG:object_",
        "target": "RETURN",
        "action": "SPLAT",
        "tags": ["HTML_ENCODED"],
        "untags": ["HTML_DECODED"],
    },
    {
        # escape_silent must  be patched in addition to escape because if the
        # C-implemented version of this is used we won't patch the internal
        # call (to the C-implemented `escape`)
        "module": "markupsafe",
        # markupsafe includes both a C implementation and a native python implementation
        # These methods accept a kwarg in the native implementation only
        "method_name": ["escape", "escape_silent"],
        "source": "ARG_0,KWARG:s",
        "target": "RETURN",
        "action": "ENCODE_HTML_SPLAT",
        "tags": ["HTML_ENCODED"],
        "untags": ["HTML_DECODED"],
    },
    {
        # markupsafe includes both a C implementation and a native python implementation",
        # soft_str() accepts a kwarg in the native implementation only",
        "module": "markupsafe",
        "method_name": "soft_str",
        "source": "ARG_0,KWARG:s",
        "target": "RETURN",
        "action": "SPLAT",
    },
    CompositeNode(
        {
            "module": "falcon.util.uri",
            "target": "RETURN",
            "action": "SPLAT",
        },
        [
            {
                "method_name": ["encode", "encode_value"],
                "source": "ARG_0,KWARG:uri",
                "tags": ["URL_ENCODED"],
            },
            {
                "method_name": "decode",
                "source": "ARG_0,KWARG:encoded_uri",
                "untags": ["URL_ENCODED"],
            },
            {
                "method_name": "unquote_string",
                "source": "ARG_0,KWARG:quoted",
                "untags": ["URL_ENCODED"],
            },
        ],
    ),
    {
        "module": "falcon.util.misc",
        "method_name": "secure_filename",
        "source": "ARG_0",
        "target": "RETURN",
        "action": "SPLAT",
        # XPATH_ENCODED means that the string is path-traversal safe, but
        # it is still UNTRUSTED because it may be dangerous for other rules
        "tags": ["XPATH_ENCODED"],
    },
    {
        "module": "werkzeug.security",
        "method_name": "safe_join",
        "source": "ALL_ARGS",
        "target": "RETURN",
        "action": "SAFE_JOIN",
        "tags": ["SAFE_PATH"],
    },
    CompositeNode(
        {
            "module": "werkzeug.utils",
            "target": "RETURN",
        },
        [
            {
                "method_name": "secure_filename",
                "source": "ARG_0,KWARG:filename",
                "action": "SPLAT",
                # XPATH_ENCODED means that the string is path-traversal safe, but
                # it is still UNTRUSTED because it may be dangerous for other rules
                "tags": ["XPATH_ENCODED"],
            },
            {
                "method_name": "escape",
                "source": "ARG_0,KWARG:s",
                "action": "ENCODE_HTML_SPLAT",
                "tags": ["HTML_ENCODED"],
                "untags": ["HTML_DECODED"],
            },
        ],
    ),
    {
        "module": "flask.helpers",
        "method_name": "safe_join",
        "source": "ALL_ARGS",
        "target": "RETURN",
        "action": "SAFE_JOIN",
        "tags": ["SAFE_PATH"],
    },
]

register_propagation_nodes(framework_propagators)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations
from collections import defaultdict
from contextlib import suppress
import time
from types import ModuleType
from collections.abc import Iterable
from contrast_vendor import importlib_metadata
import threading

from contrast.agent.scope import contrast_scope
from contrast_fireball import Library, LibraryObservation
from contrast.reporting import Reporter
from contrast.utils.patch_utils import get_loaded_modules
from contrast.utils.libraries import (
    distributions,
    get_hash,
    get_module_distribution_metadata,
    normalize_file_name,
)
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")

IMPORTABLE_FILE_SUFFIXES = (".py", ".so")
CONTRAST_DIST_NAMES = (
    "contrast-agent",
    "contrast-agent-lib",
    "contrast-fireball",
    "contrast-agent-bundle",
)


def try_distribution(name: str) -> importlib_metadata.Distribution | None:
    with suppress(importlib_metadata.PackageNotFoundError):
        return importlib_metadata.distribution(name)
    return None


CONTRAST_DIST_HASHES = tuple(
    get_hash(dist.name, dist.version)
    for name in CONTRAST_DIST_NAMES
    if (dist := try_distribution(name))
)


class LibraryReader:
    """
    LibraryReader is responsible for reading and reporting the installed libraries
    in the running environment.
    """

    def __init__(self, reporter: Reporter) -> None:
        self.reporter = reporter
        self.analysis_thread = threading.Thread(
            target=self._read_distribution_packages, daemon=True
        )

    def start(self) -> None:
        self.analysis_thread.start()

    @contrast_scope()
    def _read_distribution_packages(self) -> None:
        """
        Reads every installed distribution package and reports it as a Library.
        """
        logger.debug("Analyzing libraries...")

        libraries = self.discover_libraries()
        self._report_discovered_libraries(libraries)

        observations = self.observed_libraries()
        self._report_observed_libraries(observations)

    def discover_libraries(self) -> list[Library]:
        """
        Reads every installed distribution package and returns it as a Library.
        """
        return [
            library
            for dist in distributions()
            if (library := library_from_distribution(dist))
        ]

    def _report_discovered_libraries(self, libraries: list[Library]) -> None:
        reportable_libraries = [
            library for library in libraries if library.file not in CONTRAST_DIST_NAMES
        ]
        if not reportable_libraries:
            return

        logger.debug(
            "Reporting discovered libraries",
            count=len(reportable_libraries),
            discovered_libraries=[
                f"{lib.file} - {lib.version}" for lib in reportable_libraries
            ],
        )
        self.reporter.new_libraries(reportable_libraries)

    def observed_libraries(self) -> list[LibraryObservation]:
        """
        Reads every loaded module and returns it as a LibraryObservation.
        """
        return LibraryObservations(get_loaded_modules().values()).to_list()

    def _report_observed_libraries(
        self, observations: list[LibraryObservation]
    ) -> None:
        reportable_observations = [
            observation
            for observation in observations
            if observation.library_hash not in CONTRAST_DIST_HASHES
        ]
        if not reportable_observations:
            return

        logger.debug(
            "Reporting observed libraries",
            count=len(reportable_observations),
            used_libraries=[
                f"{obs.library_hash} - {obs.names}" for obs in reportable_observations
            ],
        )
        self.reporter.new_library_observations(reportable_observations)


def library_from_distribution(
    dist: importlib_metadata.Distribution,
) -> Library | None:
    """
    Returns a Library object from the given distribution if it is a valid library.
    Otherwise, returns None.

    A distribution is considered a valid library if it has at least one importable file
    and it is not an editable install.
    """
    metadata = dist.metadata
    name = metadata["Name"]
    version = metadata["Version"]
    dist_hash = get_hash(name, version)
    url = metadata.get("Home-page", "")

    module_files = dist.files
    if module_files is None:
        # case where the metadata files listing files (RECORD, SOURCES.txt etc..) are missing.
        return None

    if is_editable_install(name, version, module_files):
        # We decided to omit this case since whatever the package is, its probably under active development
        # and still not a library at that point in time
        return None

    importable_module_files = {
        str(f)
        for f in module_files
        if f.name.endswith(IMPORTABLE_FILE_SUFFIXES) and f.name != "setup.py"
    }
    file_count = len(importable_module_files)
    if file_count == 0:
        return None

    current_time = int(time.time() * 1000)

    return Library(
        class_count=file_count,
        internal_date=current_time,
        external_date=current_time,
        file=name,
        hash=dist_hash,
        url=url,
        version=version,
    )


def is_editable_install(
    dist_name: str, version: str, all_files: list[importlib_metadata.PackagePath]
):
    return (
        importlib_metadata.PackagePath(f"__editable__{dist_name.lower()}-{version}.pth")
        in all_files
    )


class LibraryObservations:
    """
    LibraryObservations is responsible for collecting and reporting the loaded libraries
    in the running environment.
    """

    def __init__(self, modules: Iterable[ModuleType]) -> None:
        # {hash: [file_name]}
        self._observations: dict[str, list[str]] = defaultdict(list)

        for module in modules:
            self.add_module(module)

    def add_module(self, module: ModuleType) -> None:
        if (
            (file_name := getattr(module, "__file__", None))
            and (norm_file_name := normalize_file_name(file_name))
            and (dist_meta := get_module_distribution_metadata(module))
        ):
            self._observations[
                get_hash(dist_meta["Name"], dist_meta["Version"])
            ].append(norm_file_name)

    def to_list(self) -> list[LibraryObservation]:
        return [
            LibraryObservation(library_hash=hash, names=files)
            for hash, files in self._observations.items()
        ]



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys

import contrast
from contrast.agent.scope import current_scope, set_scope

from contrast.agent.policy import patch_manager
from contrast.utils.patch_utils import (
    build_and_apply_patch,
    register_module_patcher,
    unregister_module_patcher,
    wrap_and_watermark,
)

MODULE_TO_PATCH = "concurrent.futures.thread"
CLASS_TO_PATCH = "_WorkItem"


def build__init__patch(orig_method, _):
    """
    This patch is executed when a new _WorkItem object is created. After this happens it is put on a thread safe queue.
    Worker threads in the ThreadPoolExecutor class deque and execute the call back function contained in this object.
    We attach our request context on object creation since the thread this function is running on is the same
    thread that initially served the request.
    """

    def work_item__init__patch(wrapped, instance, args, kwargs):
        ret = wrapped(*args, **kwargs)

        # Save the scope of the current thread to copy to the thread running the route function
        instance.cs__parent_scope = current_scope()
        instance.cs__parent_context = contrast.REQUEST_CONTEXT.get()

        return ret

    return wrap_and_watermark(orig_method, work_item__init__patch)


def build_run_patch(orig_method, _):
    """
    This patch is executed in the worker thread. We need to reapply the context saved in _WorkItem.__init__
    """

    def work_item_run_patch(wrapped, instance, args, kwargs):
        with contrast.lifespan(instance.cs__parent_context):
            set_scope(*instance.cs__parent_scope)

            return wrapped(*args, **kwargs)

    return wrap_and_watermark(orig_method, work_item_run_patch)


def patch_concurrent_futures_thread(module):
    cls = getattr(module, CLASS_TO_PATCH, None)
    if cls is None:
        return

    build_and_apply_patch(cls, "__init__", build__init__patch)
    build_and_apply_patch(cls, "run", build_run_patch)


def register_patches():
    register_module_patcher(patch_concurrent_futures_thread, MODULE_TO_PATCH)


def reverse_patches():
    unregister_module_patcher(MODULE_TO_PATCH)
    module = sys.modules.get(MODULE_TO_PATCH)
    if module is None:  # pragma: no cover
        return

    cls = getattr(module, CLASS_TO_PATCH)
    if cls is None:  # pragma: no cover
        return

    patch_manager.reverse_patches_by_owner(cls)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import re
from contrast.agent.protect.rule.deserialization.custom_searcher import CustomSearcher


_MODULES = [
    "os",
    "system",
    "sys",
    "subprocess",
    "__builtin__",
    "builtins",
    "globals",
    "open",
    "popen",
]

# we only need to know if the command begins with a module we care about (or "c" and
# then the module)
MODULE_REGEXES = [re.compile(rf"^c?{module}") for module in _MODULES]


class PickleSearcher(CustomSearcher):
    ID = "UD-PICKLE-1"

    NEW_LINE = "\n"
    ESCAPED_NEW_LINE = "\\n"

    def __init__(self):
        CustomSearcher.__init__(self, self.ID)

    def impact_of(self, value: str) -> int:
        impact = self.IMPACT_NONE

        split_char = (
            self.ESCAPED_NEW_LINE if self.ESCAPED_NEW_LINE in value else self.NEW_LINE
        )
        stack_commands = value.split(split_char)
        count = 0

        for command in stack_commands:
            contains_command = any(
                re.search(regex, command) for regex in MODULE_REGEXES
            )

            if contains_command:
                count += 1

                # pushing module only the stack
                if command.startswith("c"):
                    count += 1

        if count >= 3:
            impact = self.IMPACT_CRITICAL
        elif count >= 2:
            impact = self.IMPACT_HIGH

        return impact



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations
import functools
from collections import OrderedDict

from contrast.agent.assess.policy.deadzone_node import DeadZoneNode
from contrast.agent.assess.policy.propagation_node import PropagationNode
from contrast.agent.assess.policy.source_node import SourceNode
from contrast.agent.assess.rules.trigger_rule import TriggerRule
from contrast.agent.policy.patch_location_policy import PatchLocationPolicy
from contrast.agent.policy.policy_node import PolicyNode
from contrast.agent.policy.trigger_node import TriggerNode
from contrast.agent.policy.utils import generate_policy
from contrast.utils.namespace import Namespace


class _module(Namespace):
    """
    This class represents the central registry of all policy for the agent
    """

    sources: list[SourceNode] = []
    propagators: list[PropagationNode] = []
    triggers: OrderedDict[str, TriggerRule] = OrderedDict()
    providers = OrderedDict()
    deadzones: list[DeadZoneNode] = []

    # This is used to ensure that we only ever have one PatchLocationPolicy
    # instance per patch location.
    policy_by_name: OrderedDict[str, PatchLocationPolicy] = OrderedDict()
    # This is used to represent a list of PatchLocationPolicy instances per each
    # module. It is what enables the application of module-level import hooks.
    policy_by_module: OrderedDict[str, list[PatchLocationPolicy]] = OrderedDict()

    protect_policy_by_module: OrderedDict[str, list[PatchLocationPolicy]] = (
        OrderedDict()
    )


def _process_node(node: PolicyNode) -> PatchLocationPolicy:
    """
    Create patch location policy for node and update lookup tables if necessary
    """
    patch_policy = _module.policy_by_name.setdefault(
        node.name, PatchLocationPolicy(node)
    )
    patch_policy.add_node(node)

    policy_list = _module.policy_by_module.setdefault(node.module, [])
    policy_list.append(patch_policy)

    if isinstance(node, TriggerNode) and node.protect_mode:
        policy_list = _module.protect_policy_by_module.setdefault(node.module, [])
        policy_list.append(patch_policy)

    return patch_policy


def register_source_nodes(source_nodes: list) -> None:
    for node in generate_policy(SourceNode, source_nodes):
        _module.sources.append(node)
        _process_node(node)


def register_dynamic_source(
    *source_node_args,
    **source_node_kwargs,
) -> PatchLocationPolicy:
    node = SourceNode.dynamic(*source_node_args, **source_node_kwargs)
    node.validate()
    _module.sources.append(node)
    return _process_node(node)


def register_propagation_nodes(propagation_nodes: list) -> None:
    for node in generate_policy(PropagationNode, propagation_nodes):
        _module.propagators.append(node)
        _process_node(node)


def register_trigger_rule(rule: TriggerRule) -> None:
    assert rule.name not in _module.triggers
    _module.triggers[rule.name] = rule
    for node in rule.nodes:
        node.rule = rule
        _process_node(node)


def register_deadzone_nodes(deadzone_nodes: list) -> None:
    for node in generate_policy(DeadZoneNode, deadzone_nodes):
        _module.deadzones.append(node)
        _process_node(node)


def get_policy_by_name(name: str) -> PatchLocationPolicy | None:
    return _module.policy_by_name.get(name)


def get_policy(protect: bool = False) -> OrderedDict[str, list[PatchLocationPolicy]]:
    return _module.protect_policy_by_module if protect else _module.policy_by_module


def get_patch_policies(
    protect: bool = False,
) -> OrderedDict[str, list[PatchLocationPolicy]]:
    policy_patches = OrderedDict()
    for module, policy_list in get_policy(protect).items():
        if policies_with_patches := [
            policy for policy in policy_list if policy.has_patches
        ]:
            policy_patches[module] = policies_with_patches
    return policy_patches


def get_policy_by_module(
    module: str, protect: bool = False
) -> list[PatchLocationPolicy]:
    return get_policy(protect).get(module, [])


@functools.lru_cache(maxsize=1)
def get_string_method_nodes():
    return [
        node for node in get_policy_by_module("builtins") if node.class_name == "str"
    ]


# NOTE: it's possible that (some of) the methods below are only used by tests
def get_source_nodes() -> list[SourceNode]:
    return _module.sources


def get_propagation_nodes() -> list[PropagationNode]:
    return _module.propagators


def get_trigger_nodes() -> OrderedDict[str, TriggerRule]:
    return _module.triggers


def get_triggers_by_rule(rule_id: str) -> TriggerRule | None:
    return _module.triggers.get(rule_id)


def get_deadzone_nodes() -> list[DeadZoneNode]:
    return _module.deadzones



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import abc

from contrast_vendor.webob.multidict import MultiDict


class BaseResponseWrapper(metaclass=abc.ABCMeta):
    @property
    @abc.abstractmethod
    def body(self) -> bytes:
        """
        Response body as bytes
        """

    @property
    @abc.abstractmethod
    def headers(self) -> MultiDict:
        """
        Response headers as a dictionary-like object.

        This object **MUST** support normalized case-insensitive lookup.
        For example, the following should all return the same value:

        my_response_wrapper.headers['content-length']
        my_response_wrapper.headers['Content-Length']
        my_response_wrapper.headers['CONTENT-LENGTH']
        ... any other case variation

        In order to properly extract response headers for analysis, we currently expect
        this field to be a multidict that implements a method called `dict_of_lists()`.
        """

    @property
    @abc.abstractmethod
    def status_code(self) -> int:
        """
        Status code as an integer
        """



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.protect.rule.base_rule import BaseRule


class BotBlocker(BaseRule):
    RULE_NAME = "bot-blocker"

    def is_prefilter(self) -> bool:
        return self.enabled



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.dataflow_rule import DataflowRule
from contrast.agent.policy.registry import register_trigger_rule
from contrast.agent.policy.utils import CompositeNode


DISALLOWED_TAGS = [
    "CUSTOM_ENCODED_PATH_TRAVERSAL",
    "CUSTOM_ENCODED",
    "CUSTOM_VALIDATED_PATH_TRAVERSAL",
    "CUSTOM_VALIDATED",
    "LIMITED_CHARS",
    "NO_CONTROL_CHARS",
    "SAFE_PATH",
    "BASE64_ENCODED",
    "CSS_ENCODED",
    "CSV_ENCODED",
    "HTML_ENCODED",
    "JAVASCRIPT_ENCODED",
    "JAVA_ENCODED",
    "LDAP_ENCODED",
    "OS_ENCODED",
    "URL_ENCODED",
    "VBSCRIPT_ENCODED",
    "XML_ENCODED",
    "XPATH_ENCODED",
]


path_traversal_triggers = [
    {
        "module": "builtins",
        "method_name": "open",
        "source": "ARG_0,KWARG:file",
        "protect_mode": True,
    },
    CompositeNode(
        {
            "module": "os",
            "protect_mode": True,
        },
        [
            {
                "method_name": [
                    "open",
                    "unlink",
                    "remove",
                    "rmdir",
                    "chdir",
                    "chroot",
                ],
                "source": "ARG_0,KWARG:path",
            },
            {
                "method_name": [
                    "rename",
                    "link",
                    "symlink",
                    "replace",
                ],
                "source": "ARG_0,ARG_1,KWARG:src,KWARG:dst",
            },
            {
                "method_name": "renames",
                "source": "ARG_0,ARG_1,KWARG:old,KWARG:new",
            },
            {
                "method_name": [
                    "mkdir",
                    "chmod",
                    "lchmod",
                    "access",
                    "chflags",
                    "lchflags",
                    "truncate",
                    "utime",
                    "chown",
                    "lchown",
                    "setxattr",  # Linux only
                    "removexattr",  # Linux only
                ],
                "source": "ARG_0,KWARG:path",
            },
            {
                "method_name": "makedirs",
                "source": "ARG_0,KWARG:name",
            },
            {
                "method_name": ["walk", "fwalk"],
                "source": "ARG_0,KWARG:top",
            },
        ],
    ),
    CompositeNode(
        {
            "module": "shutil",
            "protect_mode": True,
        },
        [
            {
                "method_name": "rmtree",
                "source": "ARG_0,KWARG:path",
            },
            {
                "method_name": [
                    "copyfile",
                    "copymode",
                    "copystat",
                    "copy",
                    "copy2",
                    "copytree",
                    "move",
                ],
                "source": "ARG_0,ARG_1,KWARG:src,KWARG:dst",
            },
            {
                "method_name": "chown",
                "source": "ARG_0,KWARG:path",
            },
            {
                "method_name": "make_archive",
                "source": "ARG_0,ARG_2,ARG_3,KWARG:base_name,KWARG:root_dir,KWARG:base_dir",
            },
            {
                "method_name": "unpack_archive",
                "source": "ARG_0,ARG_1,KWARG:filename,KWARG:extract_dir",
            },
        ],
    ),
    {
        "module": [f"dbm.{name}" for name in ["ndbm", "gnu"]],
        "method_name": "open",
        # This function takes no kwargs
        "source": "ARG_0",
        "protect_mode": True,
    },
    {
        "module": "dbm.dumb",
        "method_name": "open",
        "source": "ARG_0,KWARG:file",
        "protect_mode": True,
    },
]


register_trigger_rule(
    DataflowRule.from_nodes(
        "path-traversal",
        path_traversal_triggers,
        disallowed_tags=DISALLOWED_TAGS,
    )
)



# code stolen from "six"

import sys
import types
from contrast_vendor.ported_cpython_code.cgi import parse_header

# True if we are running on Python 3.
PY3 = sys.version_info[0] == 3
PY2 = sys.version_info[0] == 2

if PY3:
    string_types = str,
    integer_types = int,
    class_types = type,
    text_type = str
    long = int
else:
    string_types = basestring,
    integer_types = (int, long)
    class_types = (type, types.ClassType)
    text_type = unicode
    long = long

# TODO check if errors is ever used

def text_(s, encoding='latin-1', errors='replace'):
    if isinstance(s, bytes):
        return s.decode(encoding, errors)
    return s

def bytes_(s, encoding='latin-1', errors='replace'):
    if isinstance(s, text_type):
        return s.encode(encoding, errors)
    return s

if PY3:
    def native_(s, encoding='latin-1', errors='replace'):
        if isinstance(s, text_type):
            return s
        return str(s, encoding, errors)
else:
    def native_(s, encoding='latin-1', errors='replace'):
        if isinstance(s, text_type):
            return s.encode(encoding, errors)
        return str(s)

try:
    from queue import Queue, Empty
except ImportError:
    from Queue import Queue, Empty

try:
    from collections.abc import MutableMapping
    from collections.abc import Iterable
except ImportError:
    from collections import MutableMapping
    from collections import Iterable

if PY3:
    from urllib import parse
    urlparse = parse
    from urllib.parse import quote as url_quote
    from urllib.parse import urlencode as url_encode, quote_plus
    from urllib.request import urlopen as url_open
else:
    import urlparse
    from urllib import quote_plus
    from urllib import quote as url_quote
    from urllib import unquote as url_unquote
    from urllib import urlencode as url_encode
    from urllib2 import urlopen as url_open

if PY3: # pragma: no cover
    def reraise(exc_info):
        etype, exc, tb = exc_info
        if exc.__traceback__ is not tb:
            raise exc.with_traceback(tb)
        raise exc
else:
    exec("def reraise(exc): raise exc[0], exc[1], exc[2]")


if PY3:
    def iteritems_(d):
        return d.items()
    def itervalues_(d):
        return d.values()
else:
    def iteritems_(d):
        return d.iteritems()
    def itervalues_(d):
        return d.itervalues()


if PY3: # pragma: no cover
    def unquote(string):
        if not string:
            return b''
        res = string.split(b'%')
        if len(res) != 1:
            string = res[0]
            for item in res[1:]:
                try:
                    string += bytes([int(item[:2], 16)]) + item[2:]
                except ValueError:
                    string += b'%' + item
        return string

    def url_unquote(s):
        return unquote(s.encode('ascii')).decode('latin-1')

    def parse_qsl_text(qs, encoding='utf-8'):
        qs = qs.encode('latin-1')
        qs = qs.replace(b'+', b' ')
        pairs = [s2 for s1 in qs.split(b'&') for s2 in s1.split(b';') if s2]
        for name_value in pairs:
            nv = name_value.split(b'=', 1)
            if len(nv) != 2:
                nv.append('')
            name = unquote(nv[0])
            value = unquote(nv[1])
            yield (name.decode(encoding, errors="replace"),
                   value.decode(encoding, errors="replace"))

else:
    from urlparse import parse_qsl

    def parse_qsl_text(qs, encoding='utf-8'):
        qsl = parse_qsl(
            qs,
            keep_blank_values=True,
            strict_parsing=False
        )
        for (x, y) in qsl:
            yield (x.decode(encoding), y.decode(encoding))


if PY3:
    from html import escape
else:
    from contrast_vendor.ported_cpython_code.cgi import escape


if PY3:
    from contrast_vendor.ported_cpython_code import cgi
    import tempfile
    from contrast_vendor.ported_cpython_code.cgi import FieldStorage as _cgi_FieldStorage

    # Various different FieldStorage work-arounds required on Python 3.x
    class cgi_FieldStorage(_cgi_FieldStorage): # pragma: no cover

        # Work around https://bugs.python.org/issue27777
        def make_file(self):
            if self._binary_file or self.length >= 0:
                return tempfile.TemporaryFile("wb+")
            else:
                return tempfile.TemporaryFile(
                    "w+",
                    encoding=self.encoding, newline='\n'
                )

        # Work around http://bugs.python.org/issue23801
        # This is taken exactly from Python 3.5's cgi.py module
        def read_multi(self, environ, keep_blank_values, strict_parsing):
            """Internal: read a part that is itself multipart."""
            ib = self.innerboundary
            if not cgi.valid_boundary(ib):
                raise ValueError(
                    'Invalid boundary in multipart form: %r' % (ib,))
            self.list = []
            if self.qs_on_post:
                query = cgi.urllib.parse.parse_qsl(
                    self.qs_on_post, self.keep_blank_values,
                    self.strict_parsing,
                    encoding=self.encoding, errors=self.errors)
                for key, value in query:
                    self.list.append(cgi.MiniFieldStorage(key, value))

            klass = self.FieldStorageClass or self.__class__
            first_line = self.fp.readline()  # bytes
            if not isinstance(first_line, bytes):
                raise ValueError("%s should return bytes, got %s"
                                 % (self.fp, type(first_line).__name__))
            self.bytes_read += len(first_line)

            # Ensure that we consume the file until we've hit our innerboundary
            while (first_line.strip() != (b"--" + self.innerboundary) and
                    first_line):
                first_line = self.fp.readline()
                self.bytes_read += len(first_line)

            while True:
                parser = cgi.FeedParser()
                hdr_text = b""
                while True:
                    data = self.fp.readline()
                    hdr_text += data
                    if not data.strip():
                        break
                if not hdr_text:
                    break
                # parser takes strings, not bytes
                self.bytes_read += len(hdr_text)
                parser.feed(hdr_text.decode(self.encoding, self.errors))
                headers = parser.close()
                # Some clients add Content-Length for part headers, ignore them
                if 'content-length' in headers:
                    filename = None
                    if 'content-disposition' in self.headers:
                        cdisp, pdict = parse_header(self.headers['content-disposition'])
                        if 'filename' in pdict:
                            filename = pdict['filename']
                    if filename is None:
                        del headers['content-length']
                part = klass(self.fp, headers, ib, environ, keep_blank_values,
                             strict_parsing, self.limit-self.bytes_read,
                             self.encoding, self.errors)
                self.bytes_read += part.bytes_read
                self.list.append(part)
                if part.done or self.bytes_read >= self.length > 0:
                    break
            self.skip_lines()
else:
    from contrast_vendor.ported_cpython_code.cgi import FieldStorage as cgi_FieldStorage



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.response.base_body_only_rule import BaseBodyOnlyRule


class ParameterPollutionRule(BaseBodyOnlyRule):
    @property
    def name(self):
        return "parameter-pollution"

    def is_body_violated(self, body, form_tags):
        """
        Rule is violated if:
        1. action attribute is missing within the form tag
        2. action attribute is unset or empty
        :param body: response body
        :param form_tags: list of Tag namedtuple
        :return: bool, properties dict
        """
        if not form_tags:
            return False, {}

        for tag in form_tags:
            action_attr_found = False

            for attr_name, attr_value in tag.attrs:
                if attr_name.lower() == "action":
                    action_attr_found = True

                    if not attr_value or attr_value.isspace():
                        properties = self.build_properties(tag.tag, body)
                        return True, properties

                    break

            if not action_attr_found:
                properties = self.build_properties(tag.tag, body)
                return True, properties

        return False, {}



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from collections import defaultdict
from functools import cached_property
from itertools import chain
import contrast

from contrast.agent.policy import constants
from contrast.agent.assess.assess_exceptions import ContrastAssessException
from contrast.agent.policy.trigger_node import TriggerNode
from contrast.agent.policy.utils import generate_policy
from contrast.agent.assess.rules import build_finding, send_finding
from contrast.agent.assess.rules.base_rule import BaseRule
from contrast.utils.assess.duck_utils import safe_getattr


from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class TriggerRule(BaseRule):
    UNTRUSTED = "UNTRUSTED"

    ENCODER_START = "CUSTOM_ENCODED_"
    VALIDATOR_START = "CUSTOM_VALIDATED_"
    # If a level 1 rule comes from TeamServer, it will have the
    # tag "custom-encoder-#{name}" or "custom-validator-#{name}".
    # All rules should take this into account.
    # Additionally, if something is marked "limited-chars" it means
    # it has been properly vetted to not contain dangerous input.
    LIMITED_CHARS = "LIMITED_CHARS"
    CUSTOM_ENCODED = "CUSTOM_ENCODED"
    CUSTOM_VALIDATED = "CUSTOM_VALIDATED"

    def __init__(
        self, name: str, dataflow: bool, disallowed_tags=None, required_tags=None
    ):
        self._name = name

        self.dataflow = dataflow
        # Mapping of (module, method) -> [TriggerNode]
        self._nodes: defaultdict[tuple[str, str], list[TriggerNode]] = defaultdict(list)

        self.required_tags = []
        self.populate_tags(required_tags)

        self.disallowed_tags = []
        self.populate_disallowed(disallowed_tags)

        super().__init__()

    @property
    def name(self):
        return self._name

    @property
    def disabled(self):
        """
        Property indicating whether rule is disabled
        """
        if super().disabled:
            return True

        ctx = contrast.REQUEST_CONTEXT.get()
        if ctx is None:
            return False

        return bool(
            ctx.excluded_assess_rules and self.name in ctx.excluded_assess_rules
        )

    @cached_property
    def loud_name(self):
        return str(self.name.upper().replace("-", "_"))

    @property
    def nodes(self):
        return list(chain.from_iterable(self._nodes.values()))

    def populate_tags(self, required_tags):
        if not self.dataflow:
            return

        self.validate_tags(required_tags)

        self.required_tags = required_tags if required_tags else []
        if self.UNTRUSTED not in self.required_tags:
            self.required_tags.append(self.UNTRUSTED)

    def populate_disallowed(self, disallowed_tags):
        if not self.dataflow:
            return

        self.validate_tags(disallowed_tags)

        self.disallowed_tags = disallowed_tags if disallowed_tags else []

        self.disallowed_tags.append(self.LIMITED_CHARS)
        self.disallowed_tags.append(self.CUSTOM_ENCODED)
        self.disallowed_tags.append(self.CUSTOM_VALIDATED)
        self.disallowed_tags.append(self.ENCODER_START + self.loud_name)
        self.disallowed_tags.append(self.VALIDATOR_START + self.loud_name)

    def validate_tags(self, tags):
        if not tags:
            return

        for item in tags:
            if (
                item not in constants.VALID_TAGS
                and item not in constants.VALID_SOURCE_TAGS
            ):
                raise ContrastAssessException(
                    f"Rule {self.name} had an invalid tag. {item} is not a known value"
                )

    def _is_violated(self, node, source, **kwargs):
        return node.trigger_action.is_violated(
            source, self.required_tags, self.disallowed_tags, **kwargs
        )

    def is_violated_properties(self, node, properties):
        """
        Determine whether rule was violated based on properties
        """
        return self._is_violated(node, properties)

    def is_violated(self, node, source, **kwargs):
        """
        Determine whether rule was violated based on source object
        """
        if self.count_threshold_reached():
            return False

        # The rule is violated if the object is marked as source
        if safe_getattr(source, "cs__source", False):
            return True

        return self._is_violated(node, source, **kwargs)

    def extract_source(self, node, source, args, kwargs):
        """
        Extract source from given source string based on trigger action
        """
        return node.trigger_action.extract_source(source, args, kwargs)

    def find_trigger_nodes(self, module: str, method: str) -> list[TriggerNode]:
        """
        Find the trigger node matching the module and method for this rule
        """
        return self._nodes[(module, method)]

    def create_finding(self, properties, node, target, **kwargs):
        return build_finding(self, properties, **kwargs)

    def build_and_append_finding(
        self,
        context,
        properties: dict[str, str],
        node,
        target,
        **kwargs,
    ):
        finding = self.create_finding(
            properties,
            node,
            target,
            request=context.request.to_fireball_assess_request(),
            **kwargs,
        )

        # if after creating the finding, we discover that this finding already exists
        # in context, then do not add it and return early
        for curr_finding in context.findings:
            if finding.hash == curr_finding.hash:
                return

        logger.debug(
            "Trigger %s detected: %s triggered %s",
            node.name,
            str(id(target)),
            self.name,
        )

        self.finding_reported()
        send_finding(finding, context)

    JSON_NAME = "name"
    JSON_DISALLOWED_TAGS = "disallowed_tags"
    JSON_REQUIRED_TAGS = "required_tags"
    JSON_NODES = "nodes"
    JSON_DATAFLOW = "dataflow"

    @classmethod
    def from_nodes(
        cls, name: str, dataflow: bool, nodes, disallowed_tags=None, required_tags=None
    ):
        instance = cls(name, dataflow, disallowed_tags, required_tags)
        for node in generate_policy(
            TriggerNode,
            nodes,
            dataflow=instance.dataflow,
        ):
            instance._nodes[(node.location, node.method_name)].append(node)

        return instance

    def __repr__(self):
        return f"{self.__class__.__name__} - {self.name}"



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

# ruff: noqa: F403

from .deadzones import *
from .sources import *
from .propagators import *
from .triggers import *



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import copy

from contrast.agent.assess.policy.propagators import SplatPropagator


class EncodeHtmlSplatPropagator(SplatPropagator):
    """
    This propagator is used by sanitizers such as markupsafe.escape
    and other escape functions with functionality that does not escape
    an input string if this input string has an attribute __html__.

    Because of this, if the input string has this __html__ attr,
    we remove the HTML_ENCODED tag in this propagator as we know it will
    not be sanitized by the original function (escape or the like).

    Note that we have to store the original tags because PropagationNode instances
    are only initialized once.
    """

    def propagate(self):
        """
        Store the original tags to be able to restore them later
        and then check if the input string has an __html__ attr.
        If it does, we remove the HTML_ENCODED tag for this propagation
        cycle, but then restore all tags so it does not affect
        the next time the propagation node has to be used.
        """
        string_to_escape = self.sources[0]
        self.original_tags = copy.copy(self.node.tags)

        if hasattr(string_to_escape, "__html__"):
            self.node.tags.discard("HTML_ENCODED")

        super().propagate()

    def reset_tags(self):
        self.node.tags = self.original_tags



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations
from typing import Literal
import requests

from .base_ts_message import BaseTsServerMessage
from contrast.utils.decorators import fail_loudly


class ServerInventory(BaseTsServerMessage):
    def __init__(
        self,
        *,
        operating_system: str,
        runtime_path: str,
        runtime_version: str,
        hostname: str,
        is_kubernetes: bool,
        is_docker: bool,
        cloud_provider: Literal["aws", "azure"] | None,
        cloud_resource_id: str | None,
        process_memory_limit_bytes: int | None,
    ):
        super().__init__()

        self.base_url = f"{self.settings.api_url}/agents/v1.1/"

        self.body: dict[str, str | dict[str, int]] = {
            "operating_system": operating_system,
            "runtime_path": runtime_path,
            "runtime_version": runtime_version,
            "hostname": hostname,
        }

        if cloud_provider and cloud_resource_id:
            self.body["cloud_provider"] = cloud_provider
            self.body["cloud_resource_id"] = cloud_resource_id

        if is_kubernetes:
            self.body["is_kubernetes"] = "true"

        if is_docker:
            self.body["is_docker"] = "true"

        if process_memory_limit_bytes is not None:
            self.body["memory_metrics"] = {
                "process_memory_limit_bytes": process_memory_limit_bytes,
            }

    @property
    def name(self) -> str:
        return "server-inventory"

    @property
    def path(self) -> str:
        return "/".join(
            [
                "servers",
                self.server_name_b64,
                self.server_path_b64,
                self.server_type_b64,
                "inventory",
            ]
        )

    @property
    def request_method(self):
        return requests.post

    @fail_loudly()
    def process_response(self, response, reporting_client):
        self.process_response_code(response, reporting_client)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.policy.trigger_actions.default_action import DefaultAction
from contrast.utils.patch_utils import get_arg


class FromstringAction(DefaultAction):
    """
    Custom trigger action that implements XXE fromstring rule logic.

    For lxml.etree.fromstring(), if the parser has resolve_entities disabled, then
    `fromstring` is not vulnerable to XXE.
    """

    def is_violated(
        self, source, required_tags, disallowed_tags, orig_args=None, orig_kwargs=None
    ) -> bool:
        if not super().is_violated(source, required_tags, disallowed_tags):
            return False

        from lxml import etree

        parser = get_arg(orig_args, orig_kwargs, 1, "parser", None)
        if parser is not None and not isinstance(parser, etree.XMLParser):
            return False
        return not (
            hasattr(parser, "resolve_entities") and parser.resolve_entities is False
        )



# SPDX-License-Identifier: MIT OR Apache-2.0
# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the MIT License.  See the LICENSE file in the root of this
# repository for complete details.

"""
Global state department.  Don't reload this module or everything breaks.
"""

from __future__ import annotations

import os
import sys
import warnings

from typing import Any, Callable, Iterable, Sequence, Type, cast

from ._native import make_filtering_bound_logger
from ._output import PrintLoggerFactory
from .contextvars import merge_contextvars
from .dev import ConsoleRenderer, _has_colors, set_exc_info
from .processors import StackInfoRenderer, TimeStamper, add_log_level
from .typing import BindableLogger, Context, Processor, WrappedLogger


"""
Any changes to these defaults must be reflected in:

- `getting-started`.
- structlog.stdlib.recreate_defaults()'s docstring.
"""
_BUILTIN_DEFAULT_PROCESSORS: Sequence[Processor] = [
    merge_contextvars,
    add_log_level,
    StackInfoRenderer(),
    set_exc_info,
    TimeStamper(fmt="%Y-%m-%d %H:%M:%S", utc=False),
    ConsoleRenderer(
        colors=os.environ.get("NO_COLOR", "") == ""
        and (
            os.environ.get("FORCE_COLOR", "") != ""
            or (
                _has_colors
                and sys.stdout is not None
                and hasattr(sys.stdout, "isatty")
                and sys.stdout.isatty()
            )
        )
    ),
]
_BUILTIN_DEFAULT_CONTEXT_CLASS = cast(Type[Context], dict)
_BUILTIN_DEFAULT_WRAPPER_CLASS = make_filtering_bound_logger(0)
_BUILTIN_DEFAULT_LOGGER_FACTORY = PrintLoggerFactory()
_BUILTIN_CACHE_LOGGER_ON_FIRST_USE = False


class _Configuration:
    """
    Global defaults.
    """

    is_configured: bool = False
    default_processors: Iterable[Processor] = _BUILTIN_DEFAULT_PROCESSORS[:]
    default_context_class: type[Context] = _BUILTIN_DEFAULT_CONTEXT_CLASS
    default_wrapper_class: Any = _BUILTIN_DEFAULT_WRAPPER_CLASS
    logger_factory: Callable[..., WrappedLogger] = (
        _BUILTIN_DEFAULT_LOGGER_FACTORY
    )
    cache_logger_on_first_use: bool = _BUILTIN_CACHE_LOGGER_ON_FIRST_USE


_CONFIG = _Configuration()
"""
Global defaults used when arguments to `wrap_logger` are omitted.
"""


def is_configured() -> bool:
    """
    Return whether *structlog* has been configured.

    If `False`, *structlog* is running with builtin defaults.

    .. versionadded: 18.1.0
    """
    return _CONFIG.is_configured


def get_config() -> dict[str, Any]:
    """
    Get a dictionary with the current configuration.

    .. note::

       Changes to the returned dictionary do *not* affect *structlog*.

    .. versionadded: 18.1.0
    """
    return {
        "processors": _CONFIG.default_processors,
        "context_class": _CONFIG.default_context_class,
        "wrapper_class": _CONFIG.default_wrapper_class,
        "logger_factory": _CONFIG.logger_factory,
        "cache_logger_on_first_use": _CONFIG.cache_logger_on_first_use,
    }


def get_logger(*args: Any, **initial_values: Any) -> Any:
    """
    Convenience function that returns a logger according to configuration.

    >>> from contrast_vendor.structlog import get_logger
    >>> log = get_logger(y=23)
    >>> log.info("hello", x=42)
    y=23 x=42 event='hello'

    Args:
        args:
            *Optional* positional arguments that are passed unmodified to the
            logger factory.  Therefore it depends on the factory what they
            mean.

        initial_values: Values that are used to pre-populate your contexts.

    Returns:
        A proxy that creates a correctly configured bound logger when
        necessary. The type of that bound logger depends on your configuration
        and is `structlog.BoundLogger` by default.

    See `configuration` for details.

    If you prefer CamelCase, there's an alias for your reading pleasure:
    `structlog.getLogger`.

    .. versionadded:: 0.4.0 *args*
    """
    return wrap_logger(None, logger_factory_args=args, **initial_values)


getLogger = get_logger  # noqa: N816
"""
CamelCase alias for `structlog.get_logger`.

This function is supposed to be in every source file -- we don't want it to
stick out like a sore thumb in frameworks like Twisted or Zope.
"""


def wrap_logger(
    logger: WrappedLogger | None,
    processors: Iterable[Processor] | None = None,
    wrapper_class: type[BindableLogger] | None = None,
    context_class: type[Context] | None = None,
    cache_logger_on_first_use: bool | None = None,
    logger_factory_args: Iterable[Any] | None = None,
    **initial_values: Any,
) -> Any:
    """
    Create a new bound logger for an arbitrary *logger*.

    Default values for *processors*, *wrapper_class*, and *context_class* can
    be set using `configure`.

    If you set an attribute here, `configure` calls have *no* effect for the
    *respective* attribute.

    In other words: selective overwriting of the defaults while keeping some
    *is* possible.

    Args:
        initial_values: Values that are used to pre-populate your contexts.

        logger_factory_args:
            Values that are passed unmodified as ``*logger_factory_args`` to
            the logger factory if not `None`.

    Returns:
        A proxy that creates a correctly configured bound logger when
        necessary.

    See `configure` for the meaning of the rest of the arguments.

    .. versionadded:: 0.4.0 *logger_factory_args*
    """
    return BoundLoggerLazyProxy(
        logger,
        wrapper_class=wrapper_class,
        processors=processors,
        context_class=context_class,
        cache_logger_on_first_use=cache_logger_on_first_use,
        initial_values=initial_values,
        logger_factory_args=logger_factory_args,
    )


def configure(
    processors: Iterable[Processor] | None = None,
    wrapper_class: type[BindableLogger] | None = None,
    context_class: type[Context] | None = None,
    logger_factory: Callable[..., WrappedLogger] | None = None,
    cache_logger_on_first_use: bool | None = None,
) -> None:
    """
    Configures the **global** defaults.

    They are used if `wrap_logger` or `get_logger` are called without
    arguments.

    Can be called several times, keeping an argument at `None` leaves it
    unchanged from the current setting.

    After calling for the first time, `is_configured` starts returning `True`.

    Use `reset_defaults` to undo your changes.

    Args:
        processors: The processor chain. See :doc:`processors` for details.

        wrapper_class:
            Class to use for wrapping loggers instead of
            `structlog.BoundLogger`.  See `standard-library`, :doc:`twisted`,
            and `custom-wrappers`.

        context_class:
            Class to be used for internal context keeping. The default is a
            `dict` and since dictionaries are ordered as of Python 3.6, there's
            few reasons to change this option.

        logger_factory:
            Factory to be called to create a new logger that shall be wrapped.

        cache_logger_on_first_use:
            `wrap_logger` doesn't return an actual wrapped logger but a proxy
            that assembles one when it's first used. If this option is set to
            `True`, this assembled logger is cached. See `performance`.

    .. versionadded:: 0.3.0 *cache_logger_on_first_use*
    """
    _CONFIG.is_configured = True

    if processors is not None:
        _CONFIG.default_processors = processors
    if wrapper_class is not None:
        _CONFIG.default_wrapper_class = wrapper_class
    if context_class is not None:
        _CONFIG.default_context_class = context_class
    if logger_factory is not None:
        _CONFIG.logger_factory = logger_factory
    if cache_logger_on_first_use is not None:
        _CONFIG.cache_logger_on_first_use = cache_logger_on_first_use


def configure_once(
    processors: Iterable[Processor] | None = None,
    wrapper_class: type[BindableLogger] | None = None,
    context_class: type[Context] | None = None,
    logger_factory: Callable[..., WrappedLogger] | None = None,
    cache_logger_on_first_use: bool | None = None,
) -> None:
    """
    Configures if structlog isn't configured yet.

    It does *not* matter whether it was configured using `configure` or
    `configure_once` before.

    Raises:
        RuntimeWarning: if repeated configuration is attempted.
    """
    if not _CONFIG.is_configured:
        configure(
            processors=processors,
            wrapper_class=wrapper_class,
            context_class=context_class,
            logger_factory=logger_factory,
            cache_logger_on_first_use=cache_logger_on_first_use,
        )
    else:
        warnings.warn(
            "Repeated configuration attempted.", RuntimeWarning, stacklevel=2
        )


def reset_defaults() -> None:
    """
    Resets global default values to builtin defaults.

    `is_configured` starts returning `False` afterwards.
    """
    _CONFIG.is_configured = False
    _CONFIG.default_processors = _BUILTIN_DEFAULT_PROCESSORS[:]
    _CONFIG.default_wrapper_class = _BUILTIN_DEFAULT_WRAPPER_CLASS
    _CONFIG.default_context_class = _BUILTIN_DEFAULT_CONTEXT_CLASS
    _CONFIG.logger_factory = _BUILTIN_DEFAULT_LOGGER_FACTORY
    _CONFIG.cache_logger_on_first_use = _BUILTIN_CACHE_LOGGER_ON_FIRST_USE


class BoundLoggerLazyProxy:
    """
    Instantiates a bound logger on first usage.

    Takes both configuration and instantiation parameters into account.

    The only points where a bound logger changes state are ``bind()``,
    ``unbind()``, and ``new()`` and that return the actual ``BoundLogger``.

    If and only if configuration says so, that actual bound logger is cached on
    first usage.

    .. versionchanged:: 0.4.0 Added support for *logger_factory_args*.
    """

    # fulfill BindableLogger protocol without carrying accidental state
    @property
    def _context(self) -> dict[str, str]:
        return self._initial_values

    def __init__(
        self,
        logger: WrappedLogger | None,
        wrapper_class: type[BindableLogger] | None = None,
        processors: Iterable[Processor] | None = None,
        context_class: type[Context] | None = None,
        cache_logger_on_first_use: bool | None = None,
        initial_values: dict[str, Any] | None = None,
        logger_factory_args: Any = None,
    ) -> None:
        self._logger = logger
        self._wrapper_class = wrapper_class
        self._processors = processors
        self._context_class = context_class
        self._cache_logger_on_first_use = cache_logger_on_first_use
        self._initial_values = initial_values or {}
        self._logger_factory_args = logger_factory_args or ()

    def __repr__(self) -> str:
        return (
            f"<BoundLoggerLazyProxy(logger={self._logger!r}, wrapper_class="
            f"{self._wrapper_class!r}, processors={self._processors!r}, "
            f"context_class={self._context_class!r}, "
            f"initial_values={self._initial_values!r}, "
            f"logger_factory_args={self._logger_factory_args!r})>"
        )

    def bind(self, **new_values: Any) -> BindableLogger:
        """
        Assemble a new BoundLogger from arguments and configuration.
        """
        if self._context_class:
            ctx = self._context_class(self._initial_values)
        else:
            ctx = _CONFIG.default_context_class(self._initial_values)

        _logger = self._logger
        if not _logger:
            _logger = _CONFIG.logger_factory(*self._logger_factory_args)

        if self._processors is None:
            procs = _CONFIG.default_processors
        else:
            procs = self._processors

        cls = self._wrapper_class or _CONFIG.default_wrapper_class
        # Looks like Protocols ignore definitions of __init__ so we have to
        # silence Mypy here.
        logger = cls(
            _logger, processors=procs, context=ctx  # type: ignore[call-arg]
        )

        def finalized_bind(**new_values: Any) -> BindableLogger:
            """
            Use cached assembled logger to bind potentially new values.
            """
            if new_values:
                return logger.bind(**new_values)

            return logger

        if self._cache_logger_on_first_use is True or (
            self._cache_logger_on_first_use is None
            and _CONFIG.cache_logger_on_first_use is True
        ):
            self.bind = finalized_bind  # type: ignore[method-assign]

        return finalized_bind(**new_values)

    def unbind(self, *keys: str) -> BindableLogger:
        """
        Same as bind, except unbind *keys* first.

        In our case that could be only initial values.
        """
        return self.bind().unbind(*keys)

    def try_unbind(self, *keys: str) -> BindableLogger:
        return self.bind().try_unbind(*keys)

    def new(self, **new_values: Any) -> BindableLogger:
        """
        Clear context, then bind.
        """
        if self._context_class:
            self._context_class().clear()
        else:
            _CONFIG.default_context_class().clear()

        return self.bind(**new_values)

    def __getattr__(self, name: str) -> Any:
        """
        If a logging method if called on a lazy proxy, we have to create an
        ephemeral BoundLogger first.
        """
        if name == "__isabstractmethod__":
            raise AttributeError

        bl = self.bind()

        return getattr(bl, name)

    def __getstate__(self) -> dict[str, Any]:
        """
        Our __getattr__ magic makes this necessary.
        """
        return self.__dict__

    def __setstate__(self, state: dict[str, Any]) -> None:
        """
        Our __getattr__ magic makes this necessary.
        """
        for k, v in state.items():
            setattr(self, k, v)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys


DEFAULT_LOG_PATH = "contrast-agent.log"
DEFAULT_SECURITY_LOG_PATH = "security.log"
DEFAULT_SECURITY_LOG_LEVEL = "INFO"
DEFAULT_LOG_LEVEL = "INFO"
LOGGER_NAME = "contrast"

DEFAULT_AGENT_LOGGER_PATH = sys.stderr



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

from .config_builder import ConfigBuilder
from .config_option import ConfigOption
from contrast.agent.protect.rule.mode import Mode
from contrast.utils.configuration_utils import (
    parse_disabled_rules,
    str_to_bool,
    str_to_protect_mode_enum,
)


class Protect(ConfigBuilder):
    def __init__(self):
        super().__init__()

        self.default_options = [
            ConfigOption(
                canonical_name="protect.enable",
                default_value=False,
                type_cast=str_to_bool,
            ),
            ConfigOption(
                canonical_name="protect.probe_analysis.enable",
                default_value=True,
                type_cast=str_to_bool,
            ),
            ConfigOption(
                canonical_name="protect.samples.probed", default_value=50, type_cast=int
            ),
            ConfigOption(
                canonical_name="protect.samples.blocked",
                default_value=25,
                type_cast=int,
            ),
            ConfigOption(
                canonical_name="protect.samples.exploited",
                default_value=100,
                type_cast=int,
            ),
            ConfigOption(
                canonical_name="protect.samples.blocked_at_perimeter",
                default_value=25,
                type_cast=int,
            ),
            ConfigOption(
                canonical_name="protect.rules.bot-blocker.enable",
                default_value=False,
                type_cast=str_to_bool,
            ),
            ConfigOption(
                canonical_name="protect.rules.cmd-injection.mode",
                default_value=Mode.OFF,
                type_cast=str_to_protect_mode_enum,
            ),
            ConfigOption(
                canonical_name="protect.rules.disabled_rules",
                default_value=[],
                type_cast=parse_disabled_rules,
            ),
            ConfigOption(
                canonical_name="protect.rules.method-tampering.mode",
                default_value=Mode.OFF,
                type_cast=str_to_protect_mode_enum,
            ),
            ConfigOption(
                canonical_name="protect.rules.nosql-injection.mode",
                default_value=Mode.OFF,
                type_cast=str_to_protect_mode_enum,
            ),
            ConfigOption(
                canonical_name="protect.rules.path-traversal.mode",
                default_value=Mode.OFF,
                type_cast=str_to_protect_mode_enum,
            ),
            ConfigOption(
                canonical_name="protect.rules.reflected-xss.mode",
                default_value=Mode.OFF,
                type_cast=str_to_protect_mode_enum,
            ),
            ConfigOption(
                canonical_name="protect.rules.sql-injection.mode",
                default_value=Mode.OFF,
                type_cast=str_to_protect_mode_enum,
            ),
            ConfigOption(
                canonical_name="protect.rules.ssrf.mode",
                default_value=Mode.OFF,
                type_cast=str_to_protect_mode_enum,
            ),
            ConfigOption(
                canonical_name="protect.rules.unsafe-file-upload.mode",
                default_value=Mode.OFF,
                type_cast=str_to_protect_mode_enum,
            ),
            ConfigOption(
                canonical_name="protect.rules.untrusted-deserialization.mode",
                default_value=Mode.OFF,
                type_cast=str_to_protect_mode_enum,
            ),
            ConfigOption(
                canonical_name="protect.rules.xxe.mode",
                default_value=Mode.OFF,
                type_cast=str_to_protect_mode_enum,
            ),
        ]



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from functools import cached_property
from io import BytesIO

from contrast.agent.assess.policy.source_policy import (
    cs__apply_source,
    build_source_node,
)
from contrast.utils.decorators import fail_quietly
from contrast_vendor.webob.headers import ResponseHeaders

# note: request method is not a source, per labs
# source types translated from environ sources in environ_tracker.py
SCOPE_SOURCES = {
    "path": "URI",
    "raw_path": "URI",
    "query_string": "QUERYSTRING",
    "root_path": "OTHER",
    "server": "OTHER",
    "client": "URI",
    "headers": "HEADER",
    "cookie": "COOKIE",  # not an actual scope element; we use this for headers[cookie]
}

SOURCE_DICT = {
    "module": "ASGI.scope",
    "instance_method": False,
    "target": "RETURN",
    "policy_patch": False,
}


class ASGIRequest:
    """
    A barebones request class for ASGI applications. This class is mostly useful for
    managing the `receive` coroutine, which is how ASGI apps get the request body.
    """

    HTTP_REQUEST = "http.request"
    HTTP_DISCONNECT = "http.disconnect"

    def __init__(self, scope, original_receive):
        self.scope = scope
        self.original_receive = original_receive

        self._called = False
        self._body = None
        self._wsgi_environ = None

    async def contrast__receive(self):
        """
        This coroutine should be sent to the underlying ASGI application instead of the
        original `receive`. This allows us to capture the request body while still
        providing the ASGI app with a valid `receive`-like coroutine.
        """
        if self._called:
            # We call the real receive to exhaustion the first time `contrast__receive`
            # is called. On subsequent calls, we must defer to the behavior of the real
            # receive. Some frameworks (starlette) call `receive` in a loop for
            # streaming responses and check for the http.disconnect message.
            return await self.original_receive()

        body = await self.body()
        self._called = True
        return {
            "type": self.HTTP_REQUEST,
            "body": body,
            "more_body": False,
        }

    async def body(self):
        """
        Call the original `receive` coroutine and store the result. This coroutine can
        be called multiple times without adverse consequences.
        """
        if self._body is not None:
            return self._body

        body_parts = []
        more_body = True
        while more_body:
            event = await self.original_receive()
            if event["type"] == self.HTTP_REQUEST:
                body_parts.append(event.get("body", b""))
                more_body = event.get("more_body", False)
            else:
                more_body = False
        self._body = b"".join(body_parts)
        return self._body

    async def to_wsgi_environ(self):
        if self._wsgi_environ is not None:
            return self._wsgi_environ

        body = await self.body()
        self._wsgi_environ = _scope_to_environ(self.scope, body)
        return self._wsgi_environ


class ASGIResponse:
    """
    A Response class for ASGI applications. This class handles everything related to the
    `send` coroutine, which is how ASGI applications send response data.
    """

    HTTP_RESPONSE_START = "http.response.start"
    HTTP_RESPONSE_BODY = "http.response.body"

    def __init__(self, original_send):
        self._original_send = original_send
        self._response_started = False
        self._more_body = True
        self._status_code = None
        self._asgi_headers = []
        self._body_parts = []

    async def contrast__send(self, event):
        """
        This is a heavily simplified version of the `send` implementation from Uvicorn.

        `send` is the function that the application uses to send data back to the client
        - by replacing it, we can analyze and control the response. Instead of actually
        sending data anywhere, this method simply stores the response content on our
        ASGIResponse object.
        """
        event_type = event["type"]
        if not self._response_started and self._more_body:
            if event_type != self.HTTP_RESPONSE_START:
                raise RuntimeError(
                    f"Expected ASGI event {self.HTTP_RESPONSE_START} "
                    f"but got '{event_type}'"
                )
            self._response_started = True
            self._status_code = event["status"]
            self._asgi_headers = event.get("headers", [])
        elif self._more_body:
            if event_type != self.HTTP_RESPONSE_BODY:
                raise RuntimeError(
                    f"Expected ASGI event {self.HTTP_RESPONSE_BODY} "
                    f"but got '{event_type}'"
                )
            self._body_parts.append(event.get("body", b""))
            self._more_body = event.get("more_body", False)
        else:
            raise RuntimeError(
                f"Unexpected ASGI event '{event_type}' after response already completed"
            )
        await self._original_send(event)

    @cached_property
    def body(self):
        self._check_no_more_response_body()
        return b"".join(self._body_parts)

    @cached_property
    def headers(self):
        """
        To comply with our internal ResponseWrapper API, we need headers to be strings,
        not bytes
        """
        self._check_no_more_response_body()
        return ResponseHeaders(
            [(key.decode(), value.decode()) for key, value in self._asgi_headers]
        )

    @cached_property
    def status_code(self):
        self._check_no_more_response_body()
        return self._status_code

    def _check_no_more_response_body(self):
        """
        The wrapped ASGI app must finish calling `send` before we try to access any data
        in the response. This sanity check ensures we don't silently modify the behavior
        of the original application.
        """
        if self._more_body:
            raise RuntimeError("Wrapped ASGI app did not call `send()` to completion")


def _scope_to_environ(scope: dict, body: bytes) -> dict:
    """
    Convert an asgi `scope` into a wsgi `environ` dict

    Copied from https://github.com/django/asgiref/blob/main/asgiref/wsgi.py
    and modified to our needs
    """
    environ = {
        "REQUEST_METHOD": scope.get("method", ""),
        "SCRIPT_NAME": scope.get("root_path", "").encode("utf8").decode("latin1"),
        "PATH_INFO": scope.get("path", "").encode("utf8").decode("latin1"),
        "QUERY_STRING": scope.get("query_string", b"").decode("ascii"),
        "SERVER_PROTOCOL": f"HTTP/{scope.get('http_version', '')}",
        "wsgi.version": (1, 0),
        "wsgi.url_scheme": scope.get("scheme", "http"),
        "wsgi.input": BytesIO(body),
        "wsgi.errors": BytesIO(),
        "wsgi.multithread": True,
        "wsgi.multiprocess": True,
        "wsgi.run_once": False,
    }

    # Get server name and port - required in WSGI, not in ASGI
    if scope.get("server") is not None:
        environ["SERVER_NAME"] = scope["server"][0]
        environ["SERVER_PORT"] = str(scope["server"][1] or 80)
    else:
        environ["SERVER_NAME"] = "localhost"
        environ["SERVER_PORT"] = "80"

    if scope.get("client") is not None:
        environ["REMOTE_ADDR"] = scope["client"][0]

    # Go through headers and make them into environ entries
    for name, value in scope.get("headers", []):
        name = name.decode("latin1")
        if name == "content-length":
            corrected_name = "CONTENT_LENGTH"
        elif name == "content-type":
            corrected_name = "CONTENT_TYPE"
        else:
            corrected_name = f"HTTP_{name.upper().replace('-', '_')}"
        # HTTPbis say only ASCII chars are allowed in headers, but we latin1 just in
        # case
        value = value.decode("latin1")
        if corrected_name in environ:
            value = environ[corrected_name] + "," + value
        environ[corrected_name] = value

    return environ


@fail_quietly("Failed to track ASGI scope sources")
def track_scope_sources(context, scope):
    """
    Iterate over the ASGI scope and explicitly track all relevant values. This is
    similar to the environ tracker. This function does not track the request body, which
    comes from the receive awaitable.

    There's an unfortunate possibility of starlette's request object having already
    cached some data from the scope dict. The easiest way around this would be to find
    and clear all of the places where starlette might save these variables. This could
    be avoided by going for an approach that's closer to pure ASGI instead of relying as
    much on starlette.

    @param context: the current request context
    @param scope: the ASGI scope dict
    """
    if not context.assess_enabled:
        return

    for key, value in scope.items():
        if key in ["client", "server"]:
            for elem in value or []:
                _track_scope_item(context, scope, key, elem)
        elif key == "headers":
            for header_key, header_value in value or []:
                key = "cookie" if header_key == b"cookie" else "headers"
                _track_scope_item(context, scope, key, header_key)
                _track_scope_item(context, scope, key, header_value)
        else:
            _track_scope_item(context, scope, key, value)


def _track_scope_item(context, scope, key, value):
    # there are other elements in `scope`, but only those in SCOPE_SOURCES matter to us
    if key in SCOPE_SOURCES:
        node = build_source_node(SOURCE_DICT, key, SCOPE_SOURCES[key])
        cs__apply_source(context, node, value, scope, value, (), {}, source_name=key)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from abc import ABCMeta, abstractmethod

from contrast.agent.assess.adjusted_span import AdjustedSpan


class StringToken:
    def __init__(self, value, offset):
        self.value = value
        self.offset = offset
        self.len = len(value)

    @property
    def span(self):
        return AdjustedSpan(self.offset, self.offset + self.len)

    def __len__(self):
        return self.len

    def __repr__(self):
        reprstr = "<{}({}, {})>"
        clsname = self.__class__.__name__
        return reprstr.format(clsname, repr(self.value), self.offset)


class FormatToken:
    __metaclass__ = ABCMeta

    @property
    @abstractmethod
    def span(self):
        pass

    @abstractmethod
    def format(self, args, kwargs):
        pass

    @abstractmethod
    def get_arg(self, args, kwargs):
        pass

    @abstractmethod
    def __len__(self):
        pass
